"""extend data_use with recipients and legal_basis

Revision ID: 50b3736634d2
Revises: 1739aa4a4ab7
Create Date: 2022-02-03 18:03:16.120608

"""
import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "50b3736634d2"
down_revision = "1739aa4a4ab7"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("data_uses", sa.Column("legal_basis", sa.Text(), nullable=True))
    op.add_column(
        "data_uses",
        sa.Column("recipients", postgresql.ARRAY(sa.VARCHAR()), nullable=True),
    )
    op.add_column(
        "datasets",
        sa.Column(
            "third_country_transfers", postgresql.ARRAY(sa.VARCHAR()), nullable=True
        ),
    )
    op.add_column(
        "systems",
        sa.Column(
            "third_country_transfers", postgresql.ARRAY(sa.VARCHAR()), nullable=True
        ),
    )
    op.add_column(
        "systems", sa.Column("administrating_department", sa.String(), nullable=True)
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("data_uses", "recipients")
    op.drop_column("data_uses", "legal_basis")
    op.drop_column("systems", "third_country_transfers")
    op.drop_column("datasets", "third_country_transfers")
    op.drop_column("systems", "administrating_department")
    # ### end Alembic commands ###



import secrets
from base64 import b64decode, b64encode
from binascii import Error

import bcrypt


def decode_password(password: str) -> str:
    """Tries to decode the string as base64 encoded.

    If not successful it is assumed to be unencoded and returned as it was sent.
    """
    try:
        return b64_str_to_str(password)
    except Error:
        return password


def hash_with_salt(text: bytes, salt: bytes) -> str:
    """Hashes the text using SHA-512 with the provided salt and returns the hex string
    representation"""
    return bcrypt.hashpw(text, salt).hex()


def generate_secure_random_string(length: int) -> str:
    """Generates a securely random string using Python secrets library
    that is twice the length of the specified input"""
    return secrets.token_hex(length)


def generate_salt(encoding: str = "UTF-8") -> str:
    """Generates a salt using bcrypt and returns a string using the configured
    default encoding
    """
    return bcrypt.gensalt().decode(encoding)


def bytes_to_b64_str(bytestring: bytes, encoding: str = "UTF-8") -> str:
    """Converts random bytes into a utf-8 encoded string"""
    return b64encode(bytestring).decode(encoding)


def b64_str_to_bytes(encoded_str: str, encoding: str = "UTF-8") -> bytes:
    """Converts encoded string into bytes"""
    return b64decode(encoded_str.encode(encoding))


def b64_str_to_str(encoded_str: str, encoding: str = "UTF-8") -> str:
    """Converts encoded string into str"""
    return b64decode(encoded_str).decode(encoding)


def str_to_b64_str(string: str, encoding: str = "UTF-8") -> str:
    """Converts str into a utf-8 encoded string"""
    return b64encode(string.encode(encoding)).decode(encoding)



"""
Contains all of the generic CRUD endpoints that can be
generated programmatically for each resource.
"""
from typing import Dict, List, Tuple

from loguru import logger as log
from sqlalchemy import column
from sqlalchemy import delete as _delete
from sqlalchemy import or_
from sqlalchemy import update as _update
from sqlalchemy.dialects.postgresql import Insert as _insert
from sqlalchemy.exc import SQLAlchemyError
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.future import select

from fides.api.ctl.utils import errors
from fides.lib.db.base import Base  # type: ignore[attr-defined]


# CRUD Functions
async def create_resource(
    sql_model: Base, resource_dict: Dict, async_session: AsyncSession
) -> Base:
    """Create a resource in the database."""
    with log.contextualize(
        sql_model=sql_model.__name__, fides_key=resource_dict["fides_key"]
    ):
        try:
            await get_resource(sql_model, resource_dict["fides_key"], async_session)
        except errors.NotFoundError:
            pass
        else:
            already_exists_error = errors.AlreadyExistsError(
                sql_model.__name__, resource_dict["fides_key"]
            )
            log.bind(error=already_exists_error.detail["error"]).error(  # type: ignore[index]
                "Failed to insert resource"
            )
            raise already_exists_error

        async with async_session.begin():
            try:
                log.debug("Creating resource")
                query = _insert(sql_model).values(resource_dict)
                await async_session.execute(query)
            except SQLAlchemyError:
                sa_error = errors.QueryError()
                log.bind(error=sa_error.detail["error"]).error(  # type: ignore[index]
                    "Failed to create resource"
                )
                raise sa_error

        return await get_resource(sql_model, resource_dict["fides_key"], async_session)


async def get_resource(
    sql_model: Base, fides_key: str, async_session: AsyncSession
) -> Base:
    """
    Get a resource from the databse by its FidesKey.

    Returns a SQLAlchemy model of that resource.
    """
    with log.contextualize(sql_model=sql_model.__name__, fides_key=fides_key):
        async with async_session.begin():
            try:
                log.debug("Fetching resource")
                query = select(sql_model).where(sql_model.fides_key == fides_key)
                result = await async_session.execute(query)
            except SQLAlchemyError:
                sa_error = errors.QueryError()
                log.bind(error=sa_error.detail["error"]).error(  # type: ignore[index]
                    "Failed to fetch resource"
                )
                raise sa_error

            sql_resource = result.scalars().first()
            if sql_resource is None:
                not_found_error = errors.NotFoundError(sql_model.__name__, fides_key)
                log.bind(error=not_found_error.detail["error"]).error("Resource not found")  # type: ignore[index]
                raise not_found_error

            return sql_resource


async def list_resource(sql_model: Base, async_session: AsyncSession) -> List[Base]:
    """
    Get a list of all of the resources of this type from the database.

    Returns a list of SQLAlchemy models of that resource type.
    """
    with log.contextualize(sql_model=sql_model.__name__):
        async with async_session.begin():
            try:
                log.debug("Fetching resources")
                query = select(sql_model)
                result = await async_session.execute(query)
                sql_resources = result.scalars().all()
            except SQLAlchemyError:
                error = errors.QueryError()
                log.bind(error=error.detail["error"]).error(  # type: ignore[index]
                    "Failed to fetch resources"
                )
                raise error

            return sql_resources


async def update_resource(
    sql_model: Base, resource_dict: Dict, async_session: AsyncSession
) -> Dict:
    """Update a resource in the database by its fides_key."""

    with log.contextualize(
        sql_model=sql_model.__name__, fides_key=resource_dict["fides_key"]
    ):
        await get_resource(sql_model, resource_dict["fides_key"], async_session)

        async with async_session.begin():
            try:
                log.debug("Updating resource")
                await async_session.execute(
                    _update(sql_model)
                    .where(sql_model.fides_key == resource_dict["fides_key"])
                    .values(resource_dict)
                )
            except SQLAlchemyError:
                error = errors.QueryError()
                log.bind(error=error.detail["error"]).error(  # type: ignore[index]
                    "Failed to update resource"
                )
                raise error

        return await get_resource(sql_model, resource_dict["fides_key"], async_session)


async def upsert_resources(
    sql_model: Base, resource_dicts: List[Dict], async_session: AsyncSession
) -> Tuple[int, int]:
    """
    Insert new resources into the database. If a resource already exists,
    update it by it's fides_key.

    Returns a Tuple containing the counts of inserted and updated rows.
    """

    with log.contextualize(
        sql_model=sql_model.__name__,
        fides_keys=[resource["fides_key"] for resource in resource_dicts],
    ):

        async with async_session.begin():
            try:
                log.debug("Upserting resources")
                insert_stmt = (
                    _insert(sql_model)
                    .values(resource_dicts)
                    .returning(
                        (column("xmax") == 0),  # Row was inserted
                        (column("xmax") != 0),  # Row was updated
                    )
                )

                excluded = dict(insert_stmt.excluded.items())  # type: ignore[attr-defined]
                excluded.pop("id", None)  # If updating, don't update the "id"

                result = await async_session.execute(
                    insert_stmt.on_conflict_do_update(
                        index_elements=["fides_key"],
                        set_=excluded,
                    )
                )

                inserts, updates = 0, 0
                for xmax in result:
                    inserts += 1 if xmax[0] else 0
                    updates += 1 if xmax[1] else 0

                return (inserts, updates)

            except SQLAlchemyError:
                error = errors.QueryError()
                log.bind(error=error.detail["error"]).error(  # type: ignore[index]
                    "Failed to upsert resources"
                )
                raise error


async def delete_resource(
    sql_model: Base, fides_key: str, async_session: AsyncSession
) -> Base:
    """
    Delete a resource by its fides_key.

    If the resource has child keys referring to it, also delete those.
    """

    with log.contextualize(sql_model=sql_model.__name__, fides_key=fides_key):
        sql_resource = await get_resource(sql_model, fides_key, async_session)

        async with async_session.begin():
            try:
                if hasattr(sql_model, "parent_key"):
                    log.debug("Deleting resource and its children")
                    query = (
                        _delete(sql_model)
                        .where(
                            or_(
                                sql_model.fides_key == fides_key,
                                sql_model.fides_key.like(f"{fides_key}.%"),
                            )
                        )
                        .execution_options(synchronize_session=False)
                    )
                else:
                    log.debug("Deleting resource")
                    query = _delete(sql_model).where(sql_model.fides_key == fides_key)
                print(query)
                await async_session.execute(query)
            except SQLAlchemyError:
                error = errors.QueryError()
                log.bind(error=error.detail["error"]).error(  # type: ignore[index]
                    "Failed to delete resource"
                )
                raise error

        return sql_resource



from datetime import datetime
from typing import Optional

from fastapi import APIRouter, Depends, HTTPException, Security
from fastapi_pagination import Page, Params
from fastapi_pagination.bases import AbstractPage
from fastapi_pagination.ext.sqlalchemy import paginate
from loguru import logger
from sqlalchemy.orm import Session
from sqlalchemy_utils import escape_like
from starlette.status import (
    HTTP_200_OK,
    HTTP_201_CREATED,
    HTTP_204_NO_CONTENT,
    HTTP_400_BAD_REQUEST,
    HTTP_403_FORBIDDEN,
    HTTP_404_NOT_FOUND,
)

from fides.api.ops.util.oauth_util import verify_oauth_client
from fides.core.config import FidesConfig, get_config
from fides.lib.models.client import ClientDetail
from fides.lib.models.fides_user import FidesUser
from fides.lib.models.fides_user_permissions import FidesUserPermissions
from fides.lib.oauth.api import urn_registry as urls
from fides.lib.oauth.api.deps import get_db
from fides.lib.oauth.schemas.oauth import AccessToken
from fides.lib.oauth.schemas.user import (
    UserCreate,
    UserCreateResponse,
    UserLogin,
    UserLoginResponse,
    UserResponse,
)
from fides.lib.oauth.scopes import (
    PRIVACY_REQUEST_READ,
    USER_CREATE,
    USER_DELETE,
    USER_READ,
)

router = APIRouter()


@router.post(
    urls.USERS,
    dependencies=[Security(verify_oauth_client, scopes=[USER_CREATE])],
    status_code=HTTP_201_CREATED,
    response_model=UserCreateResponse,
)
def create_user(
    *,
    db: Session = Depends(get_db),
    user_data: UserCreate,
    config: FidesConfig = Depends(get_config),
) -> FidesUser:
    """
    Create a user given a username and password.
    If `password` is sent as a base64 encoded string, it will automatically be decoded
    server-side before being encrypted and persisted.
    If `password` is sent as a plaintext string, it will be encrypted and persisted as is.
    """

    # The root user is not stored in the database so make sure here that the user name
    # is not the same as the root user name.
    if (
        config.security.root_username
        and config.security.root_username == user_data.username
    ):
        raise HTTPException(
            status_code=HTTP_400_BAD_REQUEST, detail="Username already exists."
        )

    user = FidesUser.get_by(db, field="username", value=user_data.username)

    if user:
        raise HTTPException(
            status_code=HTTP_400_BAD_REQUEST, detail="Username already exists."
        )

    user = FidesUser.create(db=db, data=user_data.dict())
    logger.info("Created user with id: '{}'.", user.id)
    FidesUserPermissions.create(
        db=db, data={"user_id": user.id, "scopes": [PRIVACY_REQUEST_READ]}
    )
    return user


@router.delete(
    urls.USER_DETAIL,
    status_code=HTTP_204_NO_CONTENT,
    dependencies=[Security(verify_oauth_client, scopes=[USER_DELETE])],
)
def delete_user(
    *,
    client: ClientDetail = Security(
        verify_oauth_client,
        scopes=[USER_DELETE],
    ),
    db: Session = Depends(get_db),
    user_id: str,
) -> None:
    """Deletes the User and associated ClientDetail if applicable."""
    user = FidesUser.get_by(db, field="id", value=user_id)

    if not user:
        raise HTTPException(
            status_code=HTTP_404_NOT_FOUND, detail=f"No user found with id {user_id}."
        )

    logger.info("User with id {} deleted by user with id {}", user_id, client.user_id)

    user.delete(db)


@router.get(
    urls.USER_DETAIL,
    dependencies=[Security(verify_oauth_client, scopes=[USER_READ])],
    response_model=UserResponse,
)
def get_user(*, db: Session = Depends(get_db), user_id: str) -> FidesUser:
    """Returns a User based on an Id"""
    user: Optional[FidesUser] = FidesUser.get_by_key_or_id(db, data={"id": user_id})
    if user is None:
        raise HTTPException(status_code=HTTP_404_NOT_FOUND, detail="User not found")

    logger.info("Returning user with id: '{}'.", user_id)
    return user


@router.get(
    urls.USERS,
    dependencies=[Security(verify_oauth_client, scopes=[USER_READ])],
    response_model=Page[UserResponse],
)
def get_users(
    *,
    db: Session = Depends(get_db),
    params: Params = Depends(),
    username: Optional[str] = None,
) -> AbstractPage[FidesUser]:
    """Returns a paginated list of all users"""
    query = FidesUser.query(db)
    if username:
        query = query.filter(FidesUser.username.ilike(f"%{escape_like(username)}%"))

    logger.info("Returning a paginated list of users.")

    return paginate(query.order_by(FidesUser.created_at.desc()), params=params)


@router.post(
    urls.LOGIN,
    status_code=HTTP_200_OK,
    response_model=UserLoginResponse,
)
def user_login(
    *,
    db: Session = Depends(get_db),
    config: FidesConfig = Depends(get_config),
    user_data: UserLogin,
) -> UserLoginResponse:
    """Login the user by creating a client if it doesn't exist, and have that client
    generate a token."""
    user: FidesUser
    client: ClientDetail
    if (
        config.security.root_username
        and config.security.root_password
        and config.security.root_username == user_data.username
        and config.security.root_password == user_data.password
    ):
        client_check = ClientDetail.get(
            db,
            object_id=config.security.oauth_root_client_id,
            config=config,
            scopes=config.security.root_user_scopes,
        )

        if not client_check:
            raise HTTPException(
                status_code=HTTP_404_NOT_FOUND, detail="No root client found."
            )

        # We have already checked for None but mypy still complains. This prevents mypy
        # from complaining.
        client = client_check
        user = FidesUser(
            id=config.security.oauth_root_client_id,
            username=config.security.root_username,
            created_at=datetime.utcnow(),
        )
    else:
        user_check: Optional[FidesUser] = FidesUser.get_by(
            db, field="username", value=user_data.username
        )

        if not user_check:
            raise HTTPException(status_code=HTTP_404_NOT_FOUND, detail="No user found.")

        if not user_check.credentials_valid(user_data.password):
            raise HTTPException(
                status_code=HTTP_403_FORBIDDEN,
                detail="Incorrect user name or password.",
            )

        # We have already checked for None but mypy still complains. This prevents mypy
        # from complaining.
        user = user_check

        client = perform_login(
            db,
            config.security.oauth_client_id_length_bytes,
            config.security.oauth_client_secret_length_bytes,
            user,
        )

    logger.info("Creating login access token")
    access_code = client.create_access_code_jwe(config.security.app_encryption_key)
    return UserLoginResponse(
        user_data=user,
        token_data=AccessToken(access_token=access_code),
    )


def perform_login(
    db: Session,
    client_id_byte_length: int,
    client_secret_btye_length: int,
    user: FidesUser,
) -> ClientDetail:
    """Performs a login by updating the FidesUser instance and creating and returning
    an associated ClientDetail.
    """

    client = user.client
    if not client:
        logger.info("Creating client for login")
        client, _ = ClientDetail.create_client_and_secret(
            db,
            client_id_byte_length,
            client_secret_btye_length,
            scopes=user.permissions.scopes,  # type: ignore
            user_id=user.id,
        )

    user.last_login_at = datetime.utcnow()
    user.save(db)

    return client



"""
Contains all of the logic related to the database including connections, setup, teardown, etc.
"""
from os import path
from typing import Literal

from alembic import command, script
from alembic.config import Config
from alembic.runtime import migration
from loguru import logger as log
from sqlalchemy.orm import Session
from sqlalchemy_utils.functions import create_database, database_exists
from sqlalchemy_utils.types.encrypted.encrypted_type import InvalidCiphertextError

from fides.api.ctl.utils.errors import get_full_exception_name
from fides.core.config import get_config
from fides.core.utils import get_db_engine
from fides.lib.db.base import Base  # type: ignore[attr-defined]

from .seed import load_default_resources
from .session import async_session

DatabaseHealth = Literal["healthy", "unhealthy", "needs migration"]

CONFIG = get_config()


def get_alembic_config(database_url: str) -> Config:
    """
    Do lots of magic to make alembic work programmatically from the CLI.
    """

    migrations_dir = path.dirname(path.abspath(__file__))
    directory = path.join(migrations_dir, "../migrations")
    config = Config(path.join(migrations_dir, "../alembic.ini"))
    config.set_main_option("script_location", directory.replace("%", "%%"))
    config.set_main_option("sqlalchemy.url", database_url)
    return config


def upgrade_db(alembic_config: Config, revision: str = "head") -> None:
    "Upgrade the database to the specified migration revision."
    log.info("Running database migrations")
    command.upgrade(alembic_config, revision)


async def init_db(database_url: str) -> None:
    """
    Runs the migrations and creates all of the database objects.
    """
    log.info("Initializing database")
    alembic_config = get_alembic_config(database_url)
    upgrade_db(alembic_config)

    async with async_session() as session:
        await load_default_resources(session)


def create_db_if_not_exists(database_url: str) -> None:
    """
    Creates a database which does not exist already.
    """
    if not database_exists(database_url):
        create_database(database_url)


def reset_db(database_url: str) -> None:
    """
    Drops all tables/metadata from the database.
    """
    log.info("Resetting database...")
    engine = get_db_engine(database_url)
    with engine.connect() as connection:
        log.info("Dropping tables...")
        Base.metadata.drop_all(connection)

        log.info("Dropping Alembic table...")
        migration_context = migration.MigrationContext.configure(connection)
        version = migration_context._version  # pylint: disable=protected-access
        if version.exists(connection):
            version.drop(connection)
    log.info("Reset complete.")


def get_db_health(database_url: str, db: Session) -> DatabaseHealth:
    """Checks if the db is reachable and up to date in alembic migrations"""
    try:
        alembic_config = get_alembic_config(database_url)
        alembic_script_directory = script.ScriptDirectory.from_config(alembic_config)
        context = migration.MigrationContext.configure(db.connection())

        if (
            context.get_current_revision()
            != alembic_script_directory.get_current_head()
        ):
            return "needs migration"
        return "healthy"
    except Exception as error:  # pylint: disable=broad-except
        error_type = get_full_exception_name(error)
        log.error("Unable to reach the database: {}: {}", error_type, error)
        return "unhealthy"


async def configure_db(database_url: str) -> None:
    """Set up the db to be used by the app."""
    try:
        create_db_if_not_exists(database_url)
        await init_db(database_url)
    except InvalidCiphertextError as cipher_error:
        log.error(
            "Unable to configure database due to a decryption error! Check to ensure your `app_encryption_key` has not changed."
        )
        log.opt(exception=True).error(cipher_error)

    except Exception as error:  # pylint: disable=broad-except
        error_type = get_full_exception_name(error)
        log.error("Unable to configure database: {}: {}", error_type, error)
        log.opt(exception=True).error(error)



from fides.api.ops.service.connectors.sql_connector import PostgreSQLConnector


class TimescaleConnector(PostgreSQLConnector):
    """TimescaleDB Connector - This currently supports identical functionality to the PostgreSQL connector"""



from typing import Any, Dict, List, Optional

from fideslang.validation import FidesKey
from loguru import logger
from sqlalchemy.orm import Session

from fides.api.ops.common_exceptions import ConnectorNotFoundException
from fides.api.ops.graph.config import CollectionAddress
from fides.api.ops.models.connectionconfig import ConnectionConfig, ConnectionType
from fides.api.ops.models.policy import ActionType, Policy
from fides.api.ops.models.privacy_request import (
    ExecutionLog,
    ExecutionLogStatus,
    PrivacyRequest,
)
from fides.api.ops.service.connectors import (
    BaseConnector,
    BigQueryConnector,
    EmailConnector,
    FidesConnector,
    ManualConnector,
    MariaDBConnector,
    MicrosoftSQLServerConnector,
    MongoDBConnector,
    MySQLConnector,
    PostgreSQLConnector,
    RedshiftConnector,
    SaaSConnector,
    SnowflakeConnector,
    TimescaleConnector,
)
from fides.api.ops.util.cache import get_cache
from fides.api.ops.util.collection_util import Row


class Connections:
    """Temporary container for connections. This will be replaced."""

    def __init__(self) -> None:
        self.connections: Dict[str, BaseConnector] = {}

    def get_connector(self, connection_config: ConnectionConfig) -> BaseConnector:
        """Return the connector corresponding to this config. Will return the existing
        connector or create one if it does not yet exist."""
        key = connection_config.key
        if key not in self.connections:
            connector = Connections.build_connector(connection_config)
            self.connections[key] = connector
        return self.connections[key]

    @staticmethod
    def build_connector(  # pylint: disable=R0911,R0912
        connection_config: ConnectionConfig,
    ) -> BaseConnector:
        """Factory method to build the appropriately typed connector from the config."""
        if connection_config.connection_type == ConnectionType.postgres:
            return PostgreSQLConnector(connection_config)
        if connection_config.connection_type == ConnectionType.mongodb:
            return MongoDBConnector(connection_config)
        if connection_config.connection_type == ConnectionType.mysql:
            return MySQLConnector(connection_config)
        if connection_config.connection_type == ConnectionType.snowflake:
            return SnowflakeConnector(connection_config)
        if connection_config.connection_type == ConnectionType.redshift:
            return RedshiftConnector(connection_config)
        if connection_config.connection_type == ConnectionType.mssql:
            return MicrosoftSQLServerConnector(connection_config)
        if connection_config.connection_type == ConnectionType.mariadb:
            return MariaDBConnector(connection_config)
        if connection_config.connection_type == ConnectionType.bigquery:
            return BigQueryConnector(connection_config)
        if connection_config.connection_type == ConnectionType.saas:
            return SaaSConnector(connection_config)
        if connection_config.connection_type == ConnectionType.manual:
            return ManualConnector(connection_config)
        if connection_config.connection_type == ConnectionType.email:
            return EmailConnector(connection_config)
        if connection_config.connection_type == ConnectionType.timescale:
            return TimescaleConnector(connection_config)
        if connection_config.connection_type == ConnectionType.fides:
            return FidesConnector(connection_config)
        raise NotImplementedError(
            f"No connector available for {connection_config.connection_type}"
        )

    def close(self) -> None:
        """Close all held connection resources."""
        for connector in self.connections.values():
            connector.close()


class TaskResources:
    """Shared information and environment for all nodes of a given task.
    This includes
     - the privacy request
     - the policy
     - redis connection
     -  configurations to any outside resources the task will require to run
    """

    def __init__(
        self,
        request: PrivacyRequest,
        policy: Policy,
        connection_configs: List[ConnectionConfig],
        session: Session,
    ):
        self.request = request
        self.policy = policy
        self.cache = get_cache()
        # tbd populate connection configurations.
        self.connection_configs: Dict[str, ConnectionConfig] = {
            c.key: c for c in connection_configs
        }
        self.connections = Connections()
        self.session = session

    def __enter__(self) -> "TaskResources":
        """Support 'with' usage for closing resources"""
        return self

    def __exit__(self, _type: Any, value: Any, traceback: Any) -> None:
        """Support 'with' usage for closing resources"""
        self.close()

    def cache_results_with_placeholders(self, key: str, value: Any) -> None:
        """Cache raw results from node. Object will be
        stored in redis under 'PLACEHOLDER_RESULTS__PRIVACY_REQUEST_ID__TYPE__COLLECTION_ADDRESS"""
        self.cache.set_encoded_object(
            f"PLACEHOLDER_RESULTS__{self.request.id}__{key}", value
        )

    def cache_object(self, key: str, value: Any) -> None:
        """Store in cache. Object will be stored in redis under 'REQUEST_ID__TYPE__ADDRESS'"""
        self.cache.set_encoded_object(f"{self.request.id}__{key}", value)

    def get_all_cached_objects(self) -> Dict[str, Optional[List[Row]]]:
        """Retrieve the access results of all steps (cache_object)"""
        value_dict = self.cache.get_encoded_objects_by_prefix(
            f"{self.request.id}__access_request"
        )
        # extract request id to return a map of address:value
        return {k.split("__")[-1]: v for k, v in value_dict.items()}

    def cache_erasure(self, key: str, value: int) -> None:
        """Cache that a node's masking is complete. Object will be stored in redis under
        'REQUEST_ID__erasure_request__ADDRESS
        '"""
        self.cache.set_encoded_object(
            f"{self.request.id}__erasure_request__{key}", value
        )

    def get_all_cached_erasures(self) -> Dict[str, int]:
        """Retrieve which collections have been masked and their row counts(cache_erasure)"""
        value_dict = self.cache.get_encoded_objects_by_prefix(
            f"{self.request.id}__erasure_request"
        )
        # extract request id to return a map of address:value
        return {k.split("__")[-1]: v for k, v in value_dict.items()}  # type: ignore

    def write_execution_log(  # pylint: disable=too-many-arguments
        self,
        connection_key: str,
        collection_address: CollectionAddress,
        fields_affected: Any,
        action_type: ActionType,
        status: ExecutionLogStatus,
        message: str = None,
    ) -> Any:
        """Store in application db. Return the created or written-to id field value."""
        db = self.session

        ExecutionLog.create(
            db=db,
            data={
                "connection_key": connection_key,
                "dataset_name": collection_address.dataset,
                "collection_name": collection_address.collection,
                "fields_affected": fields_affected,
                "action_type": action_type,
                "status": status,
                "privacy_request_id": self.request.id,
                "message": message,
            },
        )

    def get_connector(self, key: FidesKey) -> Any:
        """Create or return the client corresponding to the given ConnectionConfig key"""
        if key in self.connection_configs:
            return self.connections.get_connector(self.connection_configs[key])
        raise ConnectorNotFoundException(f"No available connector for {key}")

    def close(self) -> None:
        """Close any held resources"""
        logger.debug("Closing all task resources for {}", self.request.id)
        self.connections.close()



"""audit log email send

Revision ID: 912d801f06c0
Revises: bde646a6f51e
Create Date: 2022-09-01 16:23:10.905356

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "912d801f06c0"
down_revision = "bde646a6f51e"
branch_labels = None
depends_on = None


def upgrade():
    op.execute("alter type auditlogaction add value 'email_sent'")


def downgrade():
    op.execute("delete from auditlog where action in ('email_sent')")
    op.execute("alter type auditlogaction rename to auditlogaction_old")
    op.execute("create type auditlogaction as enum('approved', 'denied', 'finished')")
    op.execute(
        (
            "alter table auditlog alter column action type auditlogaction using "
            "action::text::auditlogaction"
        )
    )
    op.execute("drop type auditlogaction_old")



from datetime import datetime
from enum import Enum
from typing import List, Optional, Union

from fideslang.models import Dataset
from fideslang.validation import FidesKey
from pydantic import BaseModel, Extra

from fides.api.ops.models.connectionconfig import AccessLevel, ConnectionType
from fides.api.ops.schemas.api import BulkResponse, BulkUpdateFailed
from fides.api.ops.schemas.connection_configuration import connection_secrets_schemas
from fides.api.ops.schemas.saas.saas_config import SaaSConfigBase


class CreateConnectionConfiguration(BaseModel):
    """
    Schema for creating a ConnectionConfiguration

    Note that secrets are *NOT* allowed to be supplied here.
    """

    name: str
    key: Optional[FidesKey]
    connection_type: ConnectionType
    access: AccessLevel
    disabled: Optional[bool] = False
    description: Optional[str]

    class Config:
        """Restrict adding other fields through this schema and set orm_mode to support mapping to ConnectionConfig"""

        orm_mode = True
        use_enum_values = True
        extra = Extra.forbid


class CreateConnectionConfigurationWithSecrets(CreateConnectionConfiguration):
    """Schema for creating a connection configuration including secrets."""

    secrets: Optional[connection_secrets_schemas] = None
    saas_connector_type: Optional[str]

    class Config:
        orm_mode = True
        extra = Extra.forbid


class TestStatus(Enum):
    passed = "passed"
    failed = "failed"
    untested = "untested"

    def str_to_bool(self) -> Optional[bool]:
        """Translates query param string to optional/bool value
        for filtering ConnectionConfig.last_test_succeeded field"""
        if self == self.passed:
            return True
        if self == self.failed:
            return False
        return None


class SystemType(Enum):
    saas = "saas"
    database = "database"
    manual = "manual"


class ConnectionSystemTypeMap(BaseModel):
    """
    Describes the returned schema for connection types
    """

    identifier: Union[ConnectionType, str]
    type: SystemType
    human_readable: str
    encoded_icon: Optional[str]

    class Config:
        """Use enum values and set orm mode"""

        use_enum_values = True
        orm_mode = True


class ConnectionConfigurationResponse(BaseModel):
    """
    Describes the returned schema for a ConnectionConfiguration.

    Do *NOT* add "secrets" to this schema.
    """

    name: str
    key: FidesKey
    description: Optional[str]
    connection_type: ConnectionType
    access: AccessLevel
    created_at: datetime
    updated_at: Optional[datetime]
    disabled: Optional[bool] = False
    last_test_timestamp: Optional[datetime]
    last_test_succeeded: Optional[bool]
    saas_config: Optional[SaaSConfigBase]

    class Config:
        """Set orm_mode to support mapping to ConnectionConfig"""

        orm_mode = True


class BulkPutConnectionConfiguration(BulkResponse):
    """Schema with mixed success/failure responses for Bulk Create/Update of ConnectionConfiguration responses."""

    succeeded: List[ConnectionConfigurationResponse]
    failed: List[BulkUpdateFailed]


class BulkPatchConnectionConfigurationWithSecrets(BulkResponse):
    """Schema with mixed success/failure responses for Bulk Create/Update of ConnectionConfigurationWithSecrets responses."""

    succeeded: List[ConnectionConfigurationResponse]
    failed: List[BulkUpdateFailed]


class SaasConnectionTemplateValues(BaseModel):
    """Schema with values to create both a Saas ConnectionConfig and DatasetConfig from a template"""

    name: str  # For ConnectionConfig
    key: Optional[FidesKey]  # For ConnectionConfig
    description: Optional[str]  # For ConnectionConfig
    secrets: connection_secrets_schemas  # For ConnectionConfig
    instance_key: FidesKey  # For DatasetConfig.fides_key


class SaasConnectionTemplateResponse(BaseModel):
    connection: ConnectionConfigurationResponse
    dataset: Dataset



from requests import PreparedRequest

from fides.api.ops.models.connectionconfig import ConnectionConfig
from fides.api.ops.schemas.saas.strategy_configuration import (
    BasicAuthenticationConfiguration,
)
from fides.api.ops.service.authentication.authentication_strategy import (
    AuthenticationStrategy,
)
from fides.api.ops.util.saas_util import assign_placeholders


class BasicAuthenticationStrategy(AuthenticationStrategy):
    """
    Replaces the username and password placeholders with the actual credentials
    and uses them to add a basic authentication header to the incoming request.
    """

    name = "basic"
    configuration_model = BasicAuthenticationConfiguration

    def __init__(self, configuration: BasicAuthenticationConfiguration):
        self.username = configuration.username
        self.password = configuration.password

    def add_authentication(
        self, request: PreparedRequest, connection_config: ConnectionConfig
    ) -> PreparedRequest:
        """Add basic authentication to the request"""
        secrets = connection_config.secrets

        request.prepare_auth(
            auth=(
                assign_placeholders(self.username, secrets),  # type: ignore
                assign_placeholders(self.password, secrets),  # type: ignore
            )
        )
        return request



"""Contains the scan group of the commands for fides."""

import click

from fides.cli.options import (
    aws_access_key_id_option,
    aws_region_option,
    aws_secret_access_key_option,
    connection_string_option,
    coverage_threshold_option,
    credentials_id_option,
    manifests_dir_argument,
    okta_org_url_option,
    okta_token_option,
    organization_fides_key_option,
)
from fides.cli.utils import (
    handle_aws_credentials_options,
    handle_database_credentials_options,
    handle_okta_credentials_options,
    with_analytics,
)
from fides.core import dataset as _dataset
from fides.core import system as _system


@click.group(name="scan")
@click.pass_context
def scan(ctx: click.Context) -> None:
    """
    Scan external resource coverage against fides resources
    """


@scan.group(name="dataset")
@click.pass_context
def scan_dataset(ctx: click.Context) -> None:
    """
    Scan fides Dataset resources
    """


@scan_dataset.command(name="db")
@click.pass_context
@manifests_dir_argument
@credentials_id_option
@connection_string_option
@coverage_threshold_option
@with_analytics
def scan_dataset_db(
    ctx: click.Context,
    manifests_dir: str,
    connection_string: str,
    credentials_id: str,
    coverage_threshold: int,
) -> None:
    """
    Connect to a database directly via a SQLAlchemy-style connection string and
    compare the database objects to existing datasets. Connection string can be
    supplied as an option or a credentials reference to fides config.

    If there are fields within the database that aren't listed and categorized
    within one of the datasets, this counts as lacking coverage.

    Outputs missing fields and has a non-zero exit if coverage is
    under the stated threshold.
    """
    config = ctx.obj["CONFIG"]
    actual_connection_string = handle_database_credentials_options(
        fides_config=config,
        connection_string=connection_string,
        credentials_id=credentials_id,
    )

    _dataset.scan_dataset_db(
        connection_string=actual_connection_string,
        manifest_dir=manifests_dir,
        coverage_threshold=coverage_threshold,
        url=config.cli.server_url,
        headers=config.user.auth_header,
    )


@scan.group(name="system")
@click.pass_context
def scan_system(ctx: click.Context) -> None:
    """
    Scan fides System resources
    """


@scan_system.command(name="okta")
@click.pass_context
@manifests_dir_argument
@credentials_id_option
@okta_org_url_option
@okta_token_option
@organization_fides_key_option
@coverage_threshold_option
@with_analytics
def scan_system_okta(
    ctx: click.Context,
    manifests_dir: str,
    credentials_id: str,
    org_url: str,
    token: str,
    org_key: str,
    coverage_threshold: int,
) -> None:
    """
    Scans your existing systems and compares them to found Okta applications.
    Connect to an Okta admin account by providing an organization url and
    auth token or a credentials reference to fides config. Auth token and
    organization url can also be supplied by setting environment variables
    as defined by the okta python sdk.

    Outputs missing resources and has a non-zero exit if coverage is
    under the stated threshold.
    """

    config = ctx.obj["CONFIG"]
    okta_config = handle_okta_credentials_options(
        fides_config=config, token=token, org_url=org_url, credentials_id=credentials_id
    )

    _system.scan_system_okta(
        okta_config=okta_config,
        coverage_threshold=coverage_threshold,
        organization_key=org_key,
        manifest_dir=manifests_dir,
        url=config.cli.server_url,
        headers=config.user.auth_header,
    )


@scan_system.command(name="aws")
@click.pass_context
@manifests_dir_argument
@credentials_id_option
@aws_access_key_id_option
@aws_secret_access_key_option
@aws_region_option
@organization_fides_key_option
@coverage_threshold_option
@with_analytics
def scan_system_aws(
    ctx: click.Context,
    manifests_dir: str,
    credentials_id: str,
    access_key_id: str,
    secret_access_key: str,
    region: str,
    org_key: str,
    coverage_threshold: int,
) -> None:
    """
    Connect to an aws account and compares tracked resources to existing systems.
    Credentials can be supplied as options, a credentials reference to fides
    config, or boto3 environment configuration.
    Tracked resources: [Redshift, RDS, DynamoDb, S3]

    Outputs missing resources and has a non-zero exit if coverage is
    under the stated threshold.
    """
    config = ctx.obj["CONFIG"]
    aws_config = handle_aws_credentials_options(
        fides_config=config,
        access_key_id=access_key_id,
        secret_access_key=secret_access_key,
        region=region,
        credentials_id=credentials_id,
    )

    _system.scan_system_aws(
        manifest_dir=manifests_dir,
        organization_key=org_key,
        aws_config=aws_config,
        coverage_threshold=coverage_threshold,
        url=config.cli.server_url,
        headers=config.user.auth_header,
    )



"""add privacyrequest consent preferences and consent_request.privacy_request_id


Revision ID: de456534dbda
Revises: 7e218e880eaf
Create Date: 2023-01-03 22:59:45.144538

"""
import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "de456534dbda"
down_revision = "7e218e880eaf"
branch_labels = None
depends_on = None


def upgrade():
    op.add_column(
        "privacyrequest",
        sa.Column(
            "consent_preferences",
            postgresql.JSONB(astext_type=sa.Text()),
            nullable=True,
        ),
    )

    op.add_column(
        "consentrequest", sa.Column("privacy_request_id", sa.String(), nullable=True)
    )
    op.create_foreign_key(
        None, "consentrequest", "privacyrequest", ["privacy_request_id"], ["id"]
    )


def downgrade():
    op.drop_column("privacyrequest", "consent_preferences")

    op.drop_constraint(
        "consentrequest_privacy_request_id_fkey", "consentrequest", type_="foreignkey"
    )
    op.drop_column("consentrequest", "privacy_request_id")



"""add fidesctl_meta to organization object

Revision ID: 150f234ef1de
Revises: edcd28ede1f7
Create Date: 2022-03-07 20:14:00.530408

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "150f234ef1de"
down_revision = "edcd28ede1f7"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("organizations", sa.Column("fidesctl_meta", sa.JSON(), nullable=True))
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("organizations", "fidesctl_meta")
    # ### end Alembic commands ###



"""This module handles finding and parsing fides configuration files."""

# pylint: disable=C0115,C0116, E0213

from typing import Dict, Optional

from .fides_settings import FidesSettings

ENV_PREFIX = "FIDES__USER__"


class UserSettings(FidesSettings):
    """Class used to store values from the 'user' section of the config."""

    # Auth headers are set when the CLI is initiated.
    auth_header: Optional[Dict[str, str]]
    analytics_opt_out: Optional[bool]
    encryption_key: str = "test_encryption_key"

    class Config:
        env_prefix = ENV_PREFIX



from enum import Enum as EnumType
from typing import Type

from fideslang import DEFAULT_TAXONOMY

from fides.api.ops import common_exceptions


def generate_fides_data_categories() -> Type[EnumType]:
    """Programmatically generated the DataCategory enum based on the imported Fides data."""
    FidesDataCategory = EnumType(  # type: ignore
        "FidesDataCategory",
        {cat.fides_key: cat.fides_key for cat in DEFAULT_TAXONOMY.data_category},
    )
    return FidesDataCategory


DataCategory = generate_fides_data_categories()


def _validate_data_category(data_category: str) -> str:
    """Checks that the data category passed in is currently supported."""
    valid_categories = DataCategory.__members__.keys()
    if data_category not in valid_categories:
        raise common_exceptions.DataCategoryNotSupported(
            f"The data category {data_category} is not supported."
        )
    return data_category



from typing import Any, Callable

from fastapi import APIRouter as FastAPIRouter
from fastapi.types import DecoratedCallable


class APIRouter(FastAPIRouter):
    """
    Taken from: https://github.com/tiangolo/fastapi/issues/2060#issuecomment-834868906
    """

    def api_route(
        self, path: str, *, include_in_schema: bool = True, **kwargs: Any
    ) -> Callable[[DecoratedCallable], DecoratedCallable]:
        """
        Updated api_route function that automatically configures routes to have 2 versions.
        One without a trailing slash and another with it.
        """
        if path.endswith("/"):
            path = path[:-1]

        add_path = super().api_route(
            path, include_in_schema=include_in_schema, **kwargs
        )

        alternate_path = path + "/"
        add_alternate_path = super().api_route(
            alternate_path, include_in_schema=False, **kwargs
        )

        def decorator(func: DecoratedCallable) -> DecoratedCallable:
            add_alternate_path(func)
            return add_path(func)

        return decorator



CONSENT_REQUEST_VERIFICATION_TEMPLATE = "consent_request_verification.html"
SUBJECT_IDENTITY_VERIFICATION_TEMPLATE = "subject_identity_verification.html"
EMAIL_ERASURE_REQUEST_FULFILLMENT = "message_request_email_fulfillment.html"
PRIVACY_REQUEST_RECEIPT_TEMPLATE = "privacy_request_receipt.html"
PRIVACY_REQUEST_COMPLETE_DELETION_TEMPLATE = "privacy_request_complete_deletion.html"
PRIVACY_REQUEST_COMPLETE_ACCESS_TEMPLATE = "privacy_request_complete_access.html"
PRIVACY_REQUEST_ERROR_NOTIFICATION_TEMPLATE = "privacy_request_error_notification.html"
PRIVACY_REQUEST_REVIEW_DENY_TEMPLATE = "privacy_request_review_deny.html"
PRIVACY_REQUEST_REVIEW_APPROVE_TEMPLATE = "privacy_request_review_approve.html"



from enum import Enum
from typing import Any, Dict, List

from pydantic import ValidationError

from fides.api.ops.common_exceptions import NoSuchStrategyException
from fides.api.ops.common_exceptions import ValidationError as FidesopsValidationError
from fides.api.ops.schemas.saas.strategy_configuration import StrategyConfiguration
from fides.api.ops.service.processors.post_processor_strategy.post_processor_strategy import (
    PostProcessorStrategy,
)
from fides.api.ops.service.processors.post_processor_strategy.post_processor_strategy_filter import (
    FilterPostProcessorStrategy,
)
from fides.api.ops.service.processors.post_processor_strategy.post_processor_strategy_unwrap import (
    UnwrapPostProcessorStrategy,
)


class SupportedPostProcessorStrategies(Enum):
    """
    The supported methods by which Fidesops can post-process Saas connector data.
    """

    unwrap = UnwrapPostProcessorStrategy
    filter = FilterPostProcessorStrategy

    @classmethod
    def __contains__(cls, item: str) -> bool:
        try:
            cls[item]
        except KeyError:
            return False

        return True


def get_strategy(
    strategy_name: str,
    configuration: Dict[str, Any],
) -> PostProcessorStrategy:
    """
    Returns the strategy given the name and configuration.
    Raises NoSuchStrategyException if the strategy does not exist
    """
    if not SupportedPostProcessorStrategies.__contains__(strategy_name):
        valid_strategies = ", ".join([s.name for s in SupportedPostProcessorStrategies])
        raise NoSuchStrategyException(
            f"Strategy '{strategy_name}' does not exist. Valid strategies are [{valid_strategies}]"
        )
    strategy = SupportedPostProcessorStrategies[strategy_name].value
    try:
        strategy_config: StrategyConfiguration = strategy.get_configuration_model()(
            **configuration
        )
        return strategy(configuration=strategy_config)
    except ValidationError as e:
        raise FidesopsValidationError(message=str(e))


def get_strategies() -> List[PostProcessorStrategy]:
    """Returns all supported postprocessor strategies"""
    return [e.value for e in SupportedPostProcessorStrategies]



"""add datasetconfig.ctl_dataset_id fk

Adding a FK to datasetconfig pointing to the ctl_datasets table.
Also try to automigrate datasetconfig.datasets to the ctl_datasets row

Revision ID: 216cdc7944f1
Revises: 2fb48b0e268b
Create Date: 2022-12-09 22:03:51.097585

"""
import json
import uuid
from typing import Any, Dict

import sqlalchemy as sa
from alembic import op
from fideslang.models import Dataset
from sqlalchemy import text

# revision identifiers, used by Alembic.
from sqlalchemy.exc import IntegrityError
from sqlalchemy.sql.elements import TextClause

revision = "216cdc7944f1"
down_revision = "2fb48b0e268b"
branch_labels = None
depends_on = None

AUTO_MIGRATED_STRING = "auto-migrated from datasetconfig.dataset"


def upgrade():
    # Schema migration - add a nullable datasetconfig.ctl_dataset_id field.  We will shortly make it non-nullable.
    op.add_column(
        "datasetconfig", sa.Column("ctl_dataset_id", sa.String(), nullable=True)
    )
    op.create_index(
        op.f("ix_datasetconfig_ctl_dataset_id"),
        "datasetconfig",
        ["ctl_dataset_id"],
        unique=False,
    )
    op.create_foreign_key(
        "datasetconfig_ctl_dataset_id_fkey",
        "datasetconfig",
        "ctl_datasets",
        ["ctl_dataset_id"],
        ["id"],
    )

    # Data migration - automatically try to port datasetconfig.dataset -> ctl_datasets if possible.
    bind = op.get_bind()
    existing_datasetconfigs = bind.execute(
        text("SELECT id, created_at, updated_at, dataset FROM datasetconfig;")
    )
    for row in existing_datasetconfigs:
        dataset: Dict[str, Any] = row["dataset"]
        fides_key: str = dataset["fides_key"]

        insert_into_ctl_datasets_query: TextClause = text(
            "INSERT INTO ctl_datasets (id, fides_key, organization_fides_key, name, description, meta, data_categories, "
            "collections, data_qualifier, created_at, updated_at, joint_controller, retention, fides_meta, third_country_transfers, tags) "
            "VALUES (:id, :fides_key, :organization_fides_key, :name, :description, :meta, :data_categories, :collections, "
            ":data_qualifier, :created_at, :updated_at, :joint_controller, :retention, :fides_meta, :third_country_transfers, :tags)"
        )

        new_ctl_dataset_id: str = "ctl_" + str(uuid.uuid4())
        # Stashing extra text into the "meta" column so we can use this to downgrade if needed
        appended_meta: Dict = dataset["meta"] or {}
        appended_meta["fides_source"] = AUTO_MIGRATED_STRING

        validated_dataset: Dict = Dataset(
            **dataset
        ).dict()  # Validating before we store.
        validated_dataset["id"] = new_ctl_dataset_id
        validated_dataset["fides_key"] = fides_key
        validated_dataset["collections"] = json.dumps(validated_dataset["collections"])
        validated_dataset["meta"] = json.dumps(appended_meta)
        validated_dataset["created_at"] = row["created_at"]
        validated_dataset["updated_at"] = row["updated_at"]
        validated_dataset["fides_meta"] = dataset.get("fides_meta") or dataset.get(
            "fidesops_meta"
        )

        try:
            bind.execute(
                insert_into_ctl_datasets_query,
                validated_dataset,
            )
        except IntegrityError as exc:
            raise Exception(
                f"Fides attempted to copy datasetconfig.datasets into their own ctl_datasets rows but got error: {exc}. "
                f"Adjust fides_keys in ctl_datasets table to not conflict."
            )

        update_dataset_config_query: TextClause = text(
            "UPDATE datasetconfig SET ctl_dataset_id= :new_ctl_dataset_id WHERE id= :datasetconfig_id"
        )

        bind.execute(
            update_dataset_config_query,
            {"new_ctl_dataset_id": new_ctl_dataset_id, "datasetconfig_id": row["id"]},
        )


def downgrade():
    # Reverse schema migration
    op.drop_constraint(
        "datasetconfig_ctl_dataset_id_fkey", "datasetconfig", type_="foreignkey"
    )
    op.drop_index(op.f("ix_datasetconfig_ctl_dataset_id"), table_name="datasetconfig")
    op.drop_column("datasetconfig", "ctl_dataset_id")

    # Reverse data migration: remove ctl_datasets that were automatically created by the forward migration
    bind = op.get_bind()
    remove_automigrated_ctl_datasets_query: TextClause = text(
        "DELETE FROM ctl_datasets WHERE meta->>'fides_source'= :automigration_string"
    )
    bind.execute(
        remove_automigrated_ctl_datasets_query,
        {"automigration_string": AUTO_MIGRATED_STRING},
    )



from abc import abstractmethod
from typing import Any, Dict, List, Union

from fides.api.ops.service.strategy import Strategy


class PostProcessorStrategy(Strategy):
    """Abstract base class for SaaS post processor strategies"""

    @abstractmethod
    def process(
        self, data: Any, identity_data: Dict[str, Any] = None
    ) -> Union[List[Dict[str, Any]], Dict[str, Any]]:
        """Process data from SaaS connector"""



"""Merge heads

Revision ID: f53e04e5b7f5
Revises: 155fd8e51d9d, d8f8bd52754f
Create Date: 2022-07-16 02:48:43.622664

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "f53e04e5b7f5"
down_revision = ("155fd8e51d9d", "d8f8bd52754f")
branch_labels = None
depends_on = None


def upgrade():
    pass


def downgrade():
    pass



from __future__ import annotations

import json
import os
import secrets
import zipfile
from datetime import datetime
from io import BytesIO
from typing import Any, Dict, Union

import pandas as pd
from boto3 import Session
from botocore.exceptions import ClientError, ParamValidationError
from loguru import logger

from fides.api.ops.schemas.storage.storage import (
    ResponseFormat,
    S3AuthMethod,
    StorageSecrets,
)
from fides.api.ops.util.cache import get_cache, get_encryption_cache_key
from fides.api.ops.util.encryption.aes_gcm_encryption_scheme import (
    encrypt_to_bytes_verify_secrets_length,
)
from fides.api.ops.util.storage_authenticator import get_s3_session
from fides.core.config import get_config
from fides.lib.cryptography.cryptographic_util import bytes_to_b64_str

CONFIG = get_config()
LOCAL_FIDES_UPLOAD_DIRECTORY = "fides_uploads"


def encrypt_access_request_results(data: Union[str, bytes], request_id: str) -> str:
    """Encrypt data with encryption key if provided, otherwise return unencrypted data"""
    cache = get_cache()
    encryption_cache_key = get_encryption_cache_key(
        privacy_request_id=request_id,
        encryption_attr="key",
    )
    if isinstance(data, bytes):
        data = data.decode(CONFIG.security.encoding)

    encryption_key: str | None = cache.get(encryption_cache_key)
    if not encryption_key:
        return data

    bytes_encryption_key: bytes = encryption_key.encode(
        encoding=CONFIG.security.encoding
    )
    nonce: bytes = secrets.token_bytes(CONFIG.security.aes_gcm_nonce_length)
    # b64encode the entire nonce and the encrypted message together
    return bytes_to_b64_str(
        nonce
        + encrypt_to_bytes_verify_secrets_length(data, bytes_encryption_key, nonce)
    )


def write_to_in_memory_buffer(
    resp_format: str, data: Dict[str, Any], request_id: str
) -> BytesIO:
    """Write JSON/CSV data to in-memory file-like object to be passed to S3. Encrypt data if encryption key/nonce
    has been cached for the given privacy request id

    :param resp_format: str, should be one of ResponseFormat
    :param data: Dict
    :param request_id: str, The privacy request id
    """
    logger.info("Writing data to in-memory buffer")

    if resp_format == ResponseFormat.json.value:
        json_str = json.dumps(data, indent=2, default=_handle_json_encoding)
        return BytesIO(
            encrypt_access_request_results(json_str, request_id).encode(
                CONFIG.security.encoding
            )
        )

    if resp_format == ResponseFormat.csv.value:
        zipped_csvs = BytesIO()
        with zipfile.ZipFile(zipped_csvs, "w") as f:
            for key in data:
                df = pd.json_normalize(data[key])
                buffer = BytesIO()
                df.to_csv(buffer, index=False, encoding=CONFIG.security.encoding)
                buffer.seek(0)
                f.writestr(
                    f"{key}.csv",
                    encrypt_access_request_results(buffer.getvalue(), request_id),
                )

        zipped_csvs.seek(0)
        return zipped_csvs

    raise NotImplementedError(f"No handling for response format {resp_format}.")


def create_presigned_url_for_s3(
    s3_client: Session, bucket_name: str, file_key: str
) -> str:
    """ "Generate a presigned URL to share an S3 object

    :param s3_client: s3 base client
    :param bucket_name: string
    :param file_key: string
    :return: Presigned URL as string.
    """
    response = s3_client.generate_presigned_url(
        "get_object",
        Params={"Bucket": bucket_name, "Key": file_key},
        ExpiresIn=CONFIG.security.subject_request_download_link_ttl_seconds,
    )

    # The response contains the presigned URL
    return response


def upload_to_s3(  # pylint: disable=R0913
    storage_secrets: Dict[StorageSecrets, Any],
    data: Dict,
    bucket_name: str,
    file_key: str,
    resp_format: str,
    request_id: str,
    auth_method: S3AuthMethod,
) -> str:
    """Uploads arbitrary data to s3 returned from an access request"""
    logger.info("Starting S3 Upload of {}", file_key)

    try:
        my_session = get_s3_session(auth_method, storage_secrets)
        s3_client = my_session.client("s3")

        # handles file chunking
        try:
            s3_client.upload_fileobj(
                Fileobj=write_to_in_memory_buffer(resp_format, data, request_id),
                Bucket=bucket_name,
                Key=file_key,
            )
        except Exception as e:
            logger.error("Encountered error while uploading s3 object: {}", e)
            raise e

        presigned_url: str = create_presigned_url_for_s3(
            s3_client, bucket_name, file_key
        )

        return presigned_url
    except ClientError as e:
        logger.error(
            "Encountered error while uploading and generating link for s3 object: {}", e
        )
        raise e
    except ParamValidationError as e:
        raise ValueError(f"The parameters you provided are incorrect: {e}")


def _handle_json_encoding(field: Any) -> str:
    """Specify str format for datetime objects"""
    if isinstance(field, datetime):
        return field.strftime("%Y-%m-%dT%H:%M:%S")
    return field


def upload_to_local(payload: Dict, file_key: str, request_id: str) -> str:
    """Uploads access request data to a local folder - for testing/demo purposes only"""
    if not os.path.exists(LOCAL_FIDES_UPLOAD_DIRECTORY):
        os.makedirs(LOCAL_FIDES_UPLOAD_DIRECTORY)

    filename = f"{LOCAL_FIDES_UPLOAD_DIRECTORY}/{file_key}"
    data_str: str = encrypt_access_request_results(
        json.dumps(payload, default=_handle_json_encoding), request_id
    )
    with open(filename, "w") as file:  # pylint: disable=W1514
        file.write(data_str)

    return "your local fides_uploads folder"



"""
This module contains every scope used in the application.

The format for defining a scope is:
    <SCOPE_SECTION>_<SCOPE_NAME> = "<scope_section>:<scope_name>"
    CLIENT_CREATE = "client:create"

`SCOPE_REGISTRY` is intended as a comprehensive list of all available scopes.

There is a redundant `fides/lib/oauth/scopes.py` file that should not be used.
"""

# These permissions all endpoints used by the CLI's `push` & `evaluate` commands
CLI_OBJECTS_CREATE = "cli-objects:create"
CLI_OBJECTS_READ = "cli-objects:read"
CLI_OBJECTS_UPDATE = "cli-objects:update"
CLI_OBJECTS_DELETE = "cli-objects:delete"

CLIENT_CREATE = "client:create"
CLIENT_UPDATE = "client:update"
CLIENT_READ = "client:read"
CLIENT_DELETE = "client:delete"

CONFIG_READ = "config:read"

CONNECTION_TYPE_READ = "connection_type:read"

CONNECTION_CREATE_OR_UPDATE = "connection:create_or_update"
CONNECTION_READ = "connection:read"
CONNECTION_DELETE = "connection:delete"
CONNECTION_AUTHORIZE = "connection:authorize"
SAAS_CONNECTION_INSTANTIATE = "connection:instantiate"

CONSENT_READ = "consent:read"

DATABASE_RESET = "database:reset"

DATAMAP_READ = "datamap:read"

DATASET_CREATE_OR_UPDATE = "dataset:create_or_update"
DATASET_READ = "dataset:read"
DATASET_DELETE = "dataset:delete"

ENCRYPTION_EXEC = "encryption:exec"

GENERATE_EXEC = "generate:exec"

MESSAGING_CREATE_OR_UPDATE = "messaging:create_or_update"
MESSAGING_READ = "messaging:read"
MESSAGING_DELETE = "messaging:delete"

POLICY_CREATE_OR_UPDATE = "policy:create_or_update"
POLICY_READ = "policy:read"
POLICY_DELETE = "policy:delete"

PRIVACY_REQUEST_CREATE = "privacy-request:create"
PRIVACY_REQUEST_READ = "privacy-request:read"
PRIVACY_REQUEST_DELETE = "privacy-request:delete"
PRIVACY_REQUEST_CALLBACK_RESUME = (
    "privacy-request:resume"  # User has permission to resume a privacy request
)
PRIVACY_REQUEST_NOTIFICATIONS_CREATE_OR_UPDATE = (
    "privacy-request-notifications:create_or_update"
)
PRIVACY_REQUEST_NOTIFICATIONS_READ = "privacy-request-notifications:read"
PRIVACY_REQUEST_REVIEW = "privacy-request:review"
PRIVACY_REQUEST_TRANSFER = "privacy-request:transfer"
PRIVACY_REQUEST_UPLOAD_DATA = "privacy-request:upload_data"
PRIVACY_REQUEST_VIEW_DATA = "privacy-request:view_data"

RULE_CREATE_OR_UPDATE = "rule:create_or_update"
RULE_READ = "rule:read"
RULE_DELETE = "rule:delete"

SAAS_CONFIG_CREATE_OR_UPDATE = "saas_config:create_or_update"
SAAS_CONFIG_READ = "saas_config:read"
SAAS_CONFIG_DELETE = "saas_config:delete"

SCOPE_READ = "scope:read"

STORAGE_CREATE_OR_UPDATE = "storage:create_or_update"
STORAGE_READ = "storage:read"
STORAGE_DELETE = "storage:delete"

USER_CREATE = "user:create"
USER_UPDATE = "user:update"
USER_READ = "user:read"
USER_DELETE = "user:delete"
USER_PASSWORD_RESET = "user:password-reset"

USER_PERMISSION_CREATE = "user-permission:create"
USER_PERMISSION_UPDATE = "user-permission:update"
USER_PERMISSION_READ = "user-permission:read"

VALIDATE_EXEC = "validate:exec"

WEBHOOK_CREATE_OR_UPDATE = "webhook:create_or_update"
WEBHOOK_READ = "webhook:read"
WEBHOOK_DELETE = "webhook:delete"

# Should contain all scopes listed above.
SCOPE_REGISTRY = [
    CLI_OBJECTS_CREATE,
    CLI_OBJECTS_READ,
    CLI_OBJECTS_UPDATE,
    CLI_OBJECTS_DELETE,
    CLIENT_CREATE,
    CLIENT_UPDATE,
    CLIENT_READ,
    CLIENT_DELETE,
    CONFIG_READ,
    CONNECTION_READ,
    CONNECTION_CREATE_OR_UPDATE,
    CONNECTION_DELETE,
    CONNECTION_AUTHORIZE,
    SAAS_CONNECTION_INSTANTIATE,
    CONSENT_READ,
    CONNECTION_TYPE_READ,
    DATABASE_RESET,
    DATAMAP_READ,
    DATASET_CREATE_OR_UPDATE,
    DATASET_DELETE,
    DATASET_READ,
    ENCRYPTION_EXEC,
    GENERATE_EXEC,
    POLICY_CREATE_OR_UPDATE,
    POLICY_READ,
    POLICY_DELETE,
    PRIVACY_REQUEST_CREATE,
    PRIVACY_REQUEST_REVIEW,
    PRIVACY_REQUEST_READ,
    PRIVACY_REQUEST_DELETE,
    PRIVACY_REQUEST_CALLBACK_RESUME,
    PRIVACY_REQUEST_NOTIFICATIONS_CREATE_OR_UPDATE,
    PRIVACY_REQUEST_NOTIFICATIONS_READ,
    PRIVACY_REQUEST_TRANSFER,
    PRIVACY_REQUEST_UPLOAD_DATA,
    PRIVACY_REQUEST_VIEW_DATA,
    RULE_CREATE_OR_UPDATE,
    RULE_READ,
    RULE_DELETE,
    SCOPE_READ,
    STORAGE_CREATE_OR_UPDATE,
    STORAGE_DELETE,
    STORAGE_READ,
    MESSAGING_CREATE_OR_UPDATE,
    MESSAGING_DELETE,
    MESSAGING_READ,
    WEBHOOK_CREATE_OR_UPDATE,
    WEBHOOK_READ,
    WEBHOOK_DELETE,
    SAAS_CONFIG_CREATE_OR_UPDATE,
    SAAS_CONFIG_READ,
    SAAS_CONFIG_DELETE,
    USER_CREATE,
    USER_UPDATE,
    USER_READ,
    USER_PASSWORD_RESET,
    USER_DELETE,
    USER_PERMISSION_CREATE,
    USER_PERMISSION_UPDATE,
    USER_PERMISSION_READ,
    VALIDATE_EXEC,
]



"""Contains the user command group for the fides CLI."""
import click

from fides.cli.options import (
    first_name_option,
    last_name_option,
    password_option,
    username_option,
)
from fides.core.user import create_command, get_permissions_command, login_command


@click.group(name="user")
@click.pass_context
def user(ctx: click.Context) -> None:
    """
    Click command group for interacting with user-related functionality.
    """


@user.command()
@click.pass_context
@username_option
@password_option
@first_name_option
@last_name_option
def create(
    ctx: click.Context, username: str, password: str, first_name: str, last_name: str
) -> None:
    """
    Use credentials from the credentials file to create a new user.

    Gives full permissions to the new user.
    """
    config = ctx.obj["CONFIG"]
    server_url = config.cli.server_url
    create_command(
        username=username,
        password=password,
        first_name=first_name,
        last_name=last_name,
        server_url=server_url,
    )


@user.command()
@click.pass_context
@username_option
@password_option
def login(ctx: click.Context, username: str, password: str) -> None:
    """
    Use credentials to get a user access token and write it to a credentials file.
    """
    config = ctx.obj["CONFIG"]
    server_url = config.cli.server_url
    login_command(username=username, password=password, server_url=server_url)


@user.command(name="permissions")
@click.pass_context
def get_permissions(ctx: click.Context) -> None:
    """List the scopes avaible to the current user."""
    config = ctx.obj["CONFIG"]
    server_url = config.cli.server_url
    get_permissions_command(server_url=server_url)



"""Contains the view group of the commands for fides."""

import click
from toml import dumps as toml_dumps

from fides.cli.utils import print_divider, with_analytics


@click.group(name="view")
@click.pass_context
def view(ctx: click.Context) -> None:
    """
    View various resources types.
    """


@view.command(name="config")
@click.pass_context
@click.argument("section", default="", type=str)
@click.option(
    "--exclude-unset",
    is_flag=True,
    help="Only print configuration values explicitly set by the user.",
)
@with_analytics
def view_config(
    ctx: click.Context, section: str = "", exclude_unset: bool = False
) -> None:
    """
    Prints the configuration values being used.
    """
    config = ctx.obj["CONFIG"]
    config_dict = config.dict(exclude_unset=exclude_unset)
    if section:
        config_dict = config_dict[section]

    print_divider()
    print(toml_dumps(config_dict))



"""add timescale db

Revision ID: cf88efa1ad89
Revises: 021d288d0ce3
Create Date: 2022-09-14 21:26:43.721397

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "cf88efa1ad89"
down_revision = "021d288d0ce3"
branch_labels = None
depends_on = None


def upgrade():
    # Add to ConnectionType enum
    op.execute("alter type connectiontype rename to connectiontype_old")
    op.execute(
        "create type connectiontype as enum('postgres', 'mongodb', 'mysql', 'https', 'snowflake', 'redshift', 'mssql', 'mariadb', 'bigquery', 'saas', 'manual', 'email', 'manual_webhook', 'timescale')"
    )
    op.execute(
        (
            "alter table connectionconfig alter column connection_type type connectiontype using "
            "connection_type::text::connectiontype"
        )
    )
    op.execute("drop type connectiontype_old")


def downgrade():
    # Remove 'timescale' from ConnectionType
    op.execute("alter type connectiontype rename to connectiontype_old")
    op.execute(
        "create type connectiontype as enum('postgres', 'mongodb', 'mysql', 'https', 'snowflake', 'redshift', 'mssql', 'mariadb', 'bigquery', 'saas', 'manual', 'email', 'manual_webhook')"
    )
    op.execute(
        (
            "alter table connectionconfig alter column connection_type type connectiontype using "
            "connection_type::text::connectiontype"
        )
    )
    op.execute("drop type connectiontype_old")



from typing import Any, List

from pydantic import BaseModel


class FidesopsSchema(BaseModel):
    """
    A base template for all other Fidesops Schemas to inherit from.
    """

    @classmethod
    def get_field_names(cls) -> List[str]:
        """Return a list of all field names specified on this schema"""
        return list(cls.schema().get("properties", {}).keys())

    class Config:
        """Allow ORM access on all schemas"""

        orm_mode = True


BaseSchema = FidesopsSchema


class NoValidationSchema(BaseModel):
    """A schema to be used for API documentation only, when validation is handled later in the request process,
    but we still want valid request schemas to show up in the docs."""

    @classmethod
    def validate(cls: "NoValidationSchema", value: Any) -> Any:  # type: ignore
        """Returns value exactly as it was passed in, when validation is going to be handled later."""
        return value



from abc import ABC, abstractmethod
from enum import Enum
from typing import Any, Dict, Generic, Optional, Set, Tuple, TypeVar

from bson.errors import InvalidId
from bson.objectid import ObjectId
from loguru import logger

T = TypeVar("T")


class DataTypeConverter(ABC, Generic[T]):
    """DataTypeConverters are responsible for converting types of other values into the type represented here."""

    def __init__(self, name: str, empty_val: T):
        self.name = name
        self.empty_val = empty_val

    @abstractmethod
    def to_value(self, other: Any) -> Optional[T]:
        """How to convert from another datatype value to this type. When extending DataTypeConverter this method should
            return either a T or None in every case and never raise an Exception.
        an Exception."""

    def empty_value(self) -> T:
        """A value that represents `empty` in whatever way makes sense for type T"""
        return self.empty_val

    def truncate(self, length: int, val: T) -> T:
        """Truncates value to given length"""
        logger.warning(
            "{} does not support length truncation. Using original masked value instead for update query.",
            self.name,
        )
        return val

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, DataTypeConverter):
            return False

        return self.__dict__ == other.__dict__


class NoOpTypeConverter(DataTypeConverter[Any]):
    """Placeholder No-op converter. This type is assigned to fields when type is unspecified."""

    def __init__(self) -> None:
        super().__init__(name="None", empty_val=None)

    def to_value(self, other: Any) -> Optional[Any]:
        """no-op"""
        return other

    def truncate(self, length: int, val: Any) -> Any:
        """No action"""
        return val


class StringTypeConverter(DataTypeConverter[str]):
    """String data type converter. This type just uses str() type conversion."""

    def __init__(self) -> None:
        super().__init__(name="string", empty_val="")

    def to_value(self, other: Any) -> Optional[str]:
        """Convert to str"""
        return str(other) if other else None

    def truncate(self, length: int, val: str) -> str:
        """Truncates value to given length"""
        return val[:length]


class IntTypeConverter(DataTypeConverter[int]):
    """Int data type converter. This type just uses built-in int() type conversion."""

    def __init__(self) -> None:
        super().__init__(name="integer", empty_val=0)

    def to_value(self, other: Any) -> Optional[int]:
        """Convert to int"""
        try:
            return int(other)

        except (ValueError, TypeError):
            return None


class FloatTypeConverter(DataTypeConverter[float]):
    """Float data type converter. This type just uses built-in float() type conversion."""

    def __init__(self) -> None:
        super().__init__(name="float", empty_val=0.0)

    def to_value(self, other: Any) -> Optional[float]:
        """Convert to float"""
        try:
            return float(other)

        except (ValueError, TypeError):
            return None


class BooleanTypeConverter(DataTypeConverter[bool]):
    """Boolean data type converter recognizing the strings True/False, 1,0, and booleans."""

    def __init__(self) -> None:
        super().__init__(name="boolean", empty_val=False)

    true_vals = ("True", "true", True, 1)
    false_vals = ("False", "false", False, 0)

    def to_value(self, other: Any) -> Optional[bool]:
        """Convert to bool"""
        if other in BooleanTypeConverter.true_vals:
            return True
        if other in BooleanTypeConverter.false_vals:
            return False
        return None


class ObjectIdTypeConverter(DataTypeConverter[ObjectId]):
    """ObjectId data type converter, allowing for conversions from strings only."""

    def __init__(self) -> None:
        super().__init__(
            name="object_id", empty_val=ObjectId("000000000000000000000000")
        )

    def to_value(self, other: Any) -> Optional[ObjectId]:
        """Convert to ObjectId."""
        t = type(other)
        if t == ObjectId:
            return other
        if t == str and len(other) == 24:
            try:
                return ObjectId(other)
            except (InvalidId, TypeError):
                return None
        return None

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, ObjectTypeConverter):
            return False

        return self.__dict__ == other.__dict__


class ObjectTypeConverter(DataTypeConverter[Dict[str, Any]]):
    """Json data type converter."""

    def __init__(self) -> None:
        super().__init__(name="object", empty_val={})

    def to_value(self, other: Any) -> Optional[Dict[str, Any]]:
        """Pass through dict values."""
        if isinstance(other, dict):
            return other
        return None

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, ObjectTypeConverter):
            return False

        return self.__dict__ == other.__dict__


class DataType(Enum):
    """Supported data types for data retrieval and erasure.

    This type list is based on json-schema, with some alterations:
    - mongo_object_id is added to address mongodb keys
    - the json-schema 'null' type is omitted
    """

    string = StringTypeConverter()
    integer = IntTypeConverter()
    float = FloatTypeConverter()
    boolean = BooleanTypeConverter()
    object_id = ObjectIdTypeConverter()
    object = ObjectTypeConverter()
    no_op = NoOpTypeConverter()


_data_type_names: Set[str] = {
    "string",
    "integer",
    "float",
    "boolean",
    "object_id",
    "object",
}


def is_valid_data_type(type_name: str) -> bool:
    """Is this type a valid data type identifier"""
    return type_name is None or type_name in _data_type_names


def get_data_type_converter(type_name: Optional[str]) -> DataTypeConverter:
    """Return the matching type converter. If an empty string or None is passed in
    will return the No-op converter, so the converter will never be set to 'None'.
    On an illegal key will raise a KeyError."""
    if not type_name:
        return DataType.no_op.value
    return DataType[type_name].value


def parse_data_type_string(type_string: Optional[str]) -> Tuple[Optional[str], bool]:
    """Parse the data type string. Arrays are expressed in the form 'type[]'.

    e.g.
    - 'string' -> ('string', false)
    - 'string[]' -> ('string', true)
    """
    if not type_string:
        return None, False
    idx = type_string.find("[]")
    if idx == -1:
        return type_string, False
    return type_string[:idx], True


def to_data_type_string(data_type: str, is_array: bool) -> str:
    """
    Appends [] to the data type if it is an array.
    """
    if data_type == DataType.no_op.name:
        return data_type
    return data_type + "[]" if is_array else data_type


def get_data_type(value: Any) -> Tuple[Optional[str], bool]:
    """
    Returns the simple or array type of the given value.
    """

    data_type = DataType.no_op.name
    is_array = False

    # cannot assume data type for missing or empty values
    if value in (None, "", [], {}):
        return data_type, is_array

    if isinstance(value, bool):
        data_type = DataType.boolean.name
    elif isinstance(value, int):
        data_type = DataType.integer.name
    elif isinstance(value, float):
        data_type = DataType.float.name
    elif isinstance(value, str):
        data_type = DataType.string.name
    elif isinstance(value, dict):
        data_type = DataType.object.name
    elif isinstance(value, list):
        is_array = True
        if all(isinstance(item, int) for item in value):
            data_type = DataType.integer.name
        elif all(isinstance(item, float) for item in value):
            data_type = DataType.float.name
        elif all(isinstance(item, str) for item in value):
            data_type = DataType.string.name
        elif all(isinstance(item, dict) for item in value):
            data_type = DataType.object.name
    return data_type, is_array


if __name__ == "__main__":  # pragma: no cover
    v = DataType.no_op.value
    for x in v.__dict__:
        print(x)
    print(v)



"""Add ga_client_id as a provided identity type

Revision ID: 392992c7733a
Revises: de456534dbda
Create Date: 2023-01-20 15:33:24.084841

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "392992c7733a"
down_revision = "de456534dbda"
branch_labels = None
depends_on = None


def upgrade():
    op.execute("ALTER TYPE providedidentitytype ADD VALUE 'ga_client_id'")


def downgrade():
    # rename the existing type
    op.execute("ALTER TYPE providedidentitytype RENAME TO providedidentitytype_old")

    # create the new type
    op.execute("CREATE TYPE providedidentitytype AS ENUM('email', 'phone_number')")

    # update the columns to use the new type
    op.execute(
        "ALTER TABLE providedidentity ALTER COLUMN field_name TYPE providedidentitytype USING field_name::text::providedidentitytype"
    )

    # remove the old type
    op.execute("DROP TYPE providedidentitytype_old")



from enum import Enum
from typing import Dict

from fastapi import Security

from fides.api.ctl.database import database
from fides.api.ctl.routes.util import API_PREFIX
from fides.api.ctl.utils import errors
from fides.api.ctl.utils.api_router import APIRouter
from fides.api.ops.api.v1 import scope_registry
from fides.api.ops.util.oauth_util import verify_oauth_client_cli
from fides.core.config import FidesConfig, get_config

CONFIG: FidesConfig = get_config()
router = APIRouter(prefix=API_PREFIX, tags=["Admin"])


class DBActions(str, Enum):
    "The available path parameters for the `/admin/db/{action}` endpoint."
    init = "init"
    reset = "reset"


@router.post(
    "/admin/db/{action}",
    tags=["Admin"],
    dependencies=[
        Security(verify_oauth_client_cli, scopes=[scope_registry.DATABASE_RESET])
    ],
)
async def db_action(action: DBActions) -> Dict:
    """
    Initiate one of the enumerated DBActions.
    """

    action_text = "initialized"
    if action == DBActions.reset:
        if not CONFIG.dev_mode:
            raise errors.FunctionalityNotConfigured(
                "unable to reset fides database outside of dev_mode."
            )

        database.reset_db(CONFIG.database.sync_database_uri)
        action_text = DBActions.reset
    await database.configure_db(CONFIG.database.sync_database_uri)
    return {"data": {"message": f"fides database {action_text}"}}



from typing import Dict, List, Literal, Optional

from fastapi import Depends, HTTPException, status
from loguru import logger
from pydantic import BaseModel
from redis.exceptions import ResponseError
from sqlalchemy.orm import Session

import fides
from fides.api.ctl.database.database import DatabaseHealth, get_db_health
from fides.api.ctl.utils.api_router import APIRouter
from fides.api.ops.api.deps import get_db
from fides.api.ops.common_exceptions import RedisConnectionError
from fides.api.ops.tasks import celery_app, get_worker_ids
from fides.api.ops.util.cache import get_cache
from fides.api.ops.util.logger import Pii
from fides.core.config import FidesConfig, get_config

CacheHealth = Literal["healthy", "unhealthy", "no cache configured"]


class CoreHealthCheck(BaseModel):
    """Healthcheck schema"""

    webserver: str
    version: str
    database: DatabaseHealth
    cache: CacheHealth
    workers_enabled: bool
    workers: List[Optional[str]]


CONFIG: FidesConfig = get_config()

router = APIRouter(tags=["Health"])


def get_cache_health() -> str:
    """Checks if the cache is reachable"""

    if not CONFIG.redis.enabled:
        return "no cache configured"
    try:
        get_cache()
        return "healthy"
    except (RedisConnectionError, ResponseError) as e:
        logger.error("Unable to reach cache: {}", Pii(str(e)))
        return "unhealthy"


@router.get(
    "/health",
    response_model=CoreHealthCheck,
    responses={
        status.HTTP_200_OK: {
            "content": {
                "application/json": {
                    "example": {
                        "webserver": "healthy",
                        "version": "1.0.0",
                        "database": "healthy",
                        "cache": "healthy",
                        "workers_enabled": "True",
                        "workers": ["celery@c606808353b5"],
                    }
                }
            }
        },
        status.HTTP_503_SERVICE_UNAVAILABLE: {
            "content": {
                "application/json": {
                    "example": {
                        "detail": {
                            "webserver": "healthy",
                            "version": "1.0.0",
                            "database": "unhealthy",
                            "cache": "healthy",
                            "workers_enabled": "True",
                            "workers": ["celery@c606808353b5"],
                        }
                    }
                }
            }
        },
    },
)
async def health(
    db: Session = Depends(get_db),
) -> Dict:  # Intentionally injecting the ops get_db
    """Confirm that the API is running and healthy."""
    database_health = get_db_health(CONFIG.database.sync_database_uri, db=db)
    cache_health = get_cache_health()
    response = CoreHealthCheck(
        webserver="healthy",
        version=str(fides.__version__),
        database=database_health,
        cache=cache_health,
        workers_enabled=False,
        workers=[],
    ).dict()
    fides_is_using_workers = not celery_app.conf["task_always_eager"]
    if fides_is_using_workers:
        response["workers_enabled"] = True
        response["workers"] = get_worker_ids()

    for _, value in response.items():
        if value == "unhealthy":
            raise HTTPException(
                status_code=status.HTTP_503_SERVICE_UNAVAILABLE, detail=response
            )

    return response



from json import JSONDecodeError
from typing import Any, Callable, Dict, List, Optional, Tuple, Union, cast

import pydash
from loguru import logger
from requests import Response

from fides.api.ops.common_exceptions import FidesopsException, PostProcessingException
from fides.api.ops.graph.traversal import TraversalNode
from fides.api.ops.models.connectionconfig import ConnectionConfig, ConnectionTestStatus
from fides.api.ops.models.policy import Policy
from fides.api.ops.models.privacy_request import Consent, PrivacyRequest
from fides.api.ops.schemas.limiter.rate_limit_config import RateLimitConfig
from fides.api.ops.schemas.saas.saas_config import (
    ClientConfig,
    ConsentRequestMap,
    ParamValue,
    SaaSRequest,
)
from fides.api.ops.schemas.saas.shared_schemas import SaaSRequestParams
from fides.api.ops.service.connectors.base_connector import BaseConnector
from fides.api.ops.service.connectors.saas.authenticated_client import (
    AuthenticatedClient,
)
from fides.api.ops.service.connectors.saas_query_config import SaaSQueryConfig
from fides.api.ops.service.pagination.pagination_strategy import PaginationStrategy
from fides.api.ops.service.processors.post_processor_strategy.post_processor_strategy import (
    PostProcessorStrategy,
)
from fides.api.ops.service.saas_request.saas_request_override_factory import (
    SaaSRequestOverrideFactory,
    SaaSRequestType,
)
from fides.api.ops.util.collection_util import Row
from fides.api.ops.util.saas_util import assign_placeholders, map_param_values


class SaaSConnector(BaseConnector[AuthenticatedClient]):
    """A connector type to integrate with third-party SaaS APIs"""

    def __init__(self, configuration: ConnectionConfig):
        super().__init__(configuration)
        required_saas_config = configuration.get_saas_config()
        assert required_saas_config is not None
        self.saas_config = required_saas_config
        self.endpoints = self.saas_config.top_level_endpoint_dict
        self.secrets = cast(Dict, configuration.secrets)
        self.current_collection_name: Optional[str] = None
        self.current_privacy_request: Optional[PrivacyRequest] = None
        self.current_saas_request: Optional[SaaSRequest] = None

    def query_config(self, node: TraversalNode) -> SaaSQueryConfig:
        """
        Returns the query config for a given node which includes the endpoints
        and connector param values for the current collection.
        """
        privacy_request = self.current_privacy_request
        assert privacy_request is not None
        return SaaSQueryConfig(
            node,
            self.endpoints,
            self.secrets,
            self.saas_config.data_protection_request,
            privacy_request,
        )

    def get_client_config(self) -> ClientConfig:
        """Utility method for getting client config according to the current class state"""
        saas_config_client_config = self.saas_config.client_config

        required_current_saas_request = self.current_saas_request
        assert required_current_saas_request is not None
        current_request_client_config = required_current_saas_request.client_config

        return current_request_client_config or saas_config_client_config

    def get_rate_limit_config(self) -> Optional[RateLimitConfig]:
        """Utility method for getting rate limit config according to the current class state"""
        saas_config_rate_limit_config = self.saas_config.rate_limit_config

        required_current_saas_request = self.current_saas_request
        assert required_current_saas_request is not None
        current_request_rate_limit_config = (
            required_current_saas_request.rate_limit_config
        )

        return (
            current_request_rate_limit_config or saas_config_rate_limit_config or None
        )

    def set_privacy_request_state(
        self, privacy_request: PrivacyRequest, node: TraversalNode
    ) -> None:
        """
        Sets the class state for the current privacy request
        """
        self.current_collection_name = node.address.collection
        self.current_privacy_request = privacy_request

    def set_saas_request_state(self, current_saas_request: SaaSRequest) -> None:
        """
        Sets the class state for the current saas request
        """
        self.current_saas_request = current_saas_request

    def unset_connector_state(self) -> None:
        """
        Unsets the class state. Called when privacy request execution is complete
        """
        self.current_collection_name = None
        self.current_privacy_request = None
        self.current_saas_request = None

    def test_connection(self) -> Optional[ConnectionTestStatus]:
        """Generates and executes a test connection based on the SaaS config"""
        test_request: SaaSRequest = self.saas_config.test_request
        self.set_saas_request_state(test_request)
        prepared_request = map_param_values(
            "test",
            f"{self.configuration.name}",
            test_request,
            self.secrets,
        )
        client: AuthenticatedClient = self.create_client()
        client.send(prepared_request, test_request.ignore_errors)
        self.unset_connector_state()
        return ConnectionTestStatus.succeeded

    def build_uri(self) -> str:
        """Build base URI for the given connector"""
        client_config = self.get_client_config()
        host = client_config.host
        return f"{client_config.protocol}://{assign_placeholders(host, self.secrets)}"

    def create_client(self) -> AuthenticatedClient:
        """Creates an authenticated request builder"""
        uri = self.build_uri()
        client_config = self.get_client_config()
        rate_limit_config = self.get_rate_limit_config()

        logger.info("Creating client to {}", uri)
        return AuthenticatedClient(
            uri, self.configuration, client_config, rate_limit_config
        )

    def retrieve_data(
        self,
        node: TraversalNode,
        policy: Policy,
        privacy_request: PrivacyRequest,
        input_data: Dict[str, List[Any]],
    ) -> List[Row]:
        """Retrieve data from SaaS APIs"""
        self.set_privacy_request_state(privacy_request, node)
        query_config: SaaSQueryConfig = self.query_config(node)

        # generate initial set of requests if read request is defined, otherwise raise an exception

        # An endpoint can be defined with multiple 'read' requests if the data for a single
        # collection can be accessed in multiple ways for example:
        #
        # 1) If a collection can be retrieved by using different identities such as email or phone number
        # 2) The complete set of results for a collection is made up of subsets. For example, to retrieve all tickets
        #    we must change a 'status' query param from 'active' to 'pending' and finally 'closed'
        read_requests: List[SaaSRequest] = query_config.get_read_requests_by_identity()
        delete_request: Optional[
            SaaSRequest
        ] = query_config.get_erasure_request_by_action("delete")

        if not read_requests:
            # if a delete request is specified for this endpoint without a read request
            # then we return a single empty row to still trigger the mask_data method
            if delete_request:
                logger.info(
                    "Skipping read for the '{}' collection, it is delete-only",
                    self.current_collection_name,
                )
                return [{}]

            raise FidesopsException(
                f"The 'read' action is not defined for the '{self.current_collection_name}' "
                f"endpoint in {self.saas_config.fides_key}"
            )

        rows: List[Row] = []
        for read_request in read_requests:
            self.set_saas_request_state(read_request)
            # check all the values specified by param_values are provided in input_data
            if self._missing_dataset_reference_values(
                input_data, read_request.param_values
            ):
                return []

            # hook for user-provided request override functions
            if read_request.request_override:
                return self._invoke_read_request_override(
                    read_request.request_override,
                    self.create_client(),
                    policy,
                    privacy_request,
                    node,
                    input_data,
                    self.secrets,
                )

            prepared_requests: List[SaaSRequestParams] = query_config.generate_requests(
                input_data, policy, read_request
            )

            # Iterates through initial list of prepared requests and through subsequent
            # requests generated by pagination. The results are added to the output
            # list of rows after each request.
            for next_request in prepared_requests:
                while next_request:
                    processed_rows, next_request = self.execute_prepared_request(  # type: ignore
                        next_request,
                        privacy_request.get_cached_identity_data(),
                        read_request,
                    )
                    rows.extend(processed_rows)
        self.unset_connector_state()
        return rows

    def _missing_dataset_reference_values(
        self, input_data: Dict[str, Any], param_values: Optional[List[ParamValue]]
    ) -> List[str]:
        """Return a list of dataset reference values that are not found in the input_data map"""

        # get the list of param_value references
        required_param_value_references = [
            param_value.name
            for param_value in param_values or []
            if param_value.references
        ]

        # extract the keys from inside the fides_grouped_inputs and append them the other input_data keys
        provided_input_keys = (
            list(input_data.get("fidesops_grouped_inputs")[0].keys())  # type: ignore
            if input_data.get("fidesops_grouped_inputs")
            else []
        ) + list(input_data.keys())

        # find the missing values
        missing_dataset_reference_values = list(
            set(required_param_value_references) - set(provided_input_keys)
        )

        if missing_dataset_reference_values:
            logger.info(
                "The '{}' request of {} is missing the following dataset reference values [{}], skipping traversal",
                self.current_collection_name,
                self.saas_config.fides_key,  # type: ignore
                ", ".join(missing_dataset_reference_values),
            )
        return missing_dataset_reference_values

    def execute_prepared_request(
        self,
        prepared_request: SaaSRequestParams,
        identity_data: Dict[str, Any],
        saas_request: SaaSRequest,
    ) -> Tuple[List[Row], Optional[SaaSRequestParams]]:
        """
        Executes the prepared request and handles response postprocessing and pagination.
        Returns processed data and request_params for next page of data if available.
        """

        client: AuthenticatedClient = self.create_client()
        response: Response = client.send(prepared_request, saas_request.ignore_errors)
        response = self._handle_errored_response(saas_request, response)
        response_data = self._unwrap_response_data(saas_request, response)

        # process response and add to rows
        rows = self.process_response_data(
            response_data,
            identity_data,
            cast(Optional[List[PostProcessorStrategy]], saas_request.postprocessors),
        )

        logger.info(
            "{} row(s) returned after postprocessing '{}' collection.",
            len(rows),
            self.current_collection_name,
        )

        # use the pagination strategy (if available) to get the next request
        next_request = None
        if saas_request.pagination:
            strategy: PaginationStrategy = PaginationStrategy.get_strategy(
                saas_request.pagination.strategy,
                saas_request.pagination.configuration,
            )
            next_request = strategy.get_next_request(
                prepared_request, self.secrets, response, saas_request.data_path
            )

        if next_request:
            logger.info(
                "Using '{}' pagination strategy to get next page for '{}'.",
                saas_request.pagination.strategy,  # type: ignore
                self.current_collection_name,
            )

        return rows, next_request

    def process_response_data(
        self,
        response_data: Union[List[Dict[str, Any]], Dict[str, Any]],
        identity_data: Dict[str, Any],
        postprocessors: Optional[List[PostProcessorStrategy]],
    ) -> List[Row]:
        """
        Runs the raw response through all available postprocessors for the request,
        forwarding the output of one postprocessor into the input of the next.

        The final result is returned as a list of processed objects.
        """

        rows: List[Row] = []
        processed_data = response_data
        for postprocessor in postprocessors or []:
            strategy: PostProcessorStrategy = PostProcessorStrategy.get_strategy(
                postprocessor.strategy, postprocessor.configuration  # type: ignore
            )
            logger.info(
                "Starting postprocessing of '{}' collection with '{}' strategy.",
                self.current_collection_name,
                postprocessor.strategy,  # type: ignore
            )
            try:
                processed_data = strategy.process(processed_data, identity_data)
            except Exception as exc:
                raise PostProcessingException(
                    f"Exception occurred during the '{postprocessor.strategy}' postprocessor "  # type: ignore
                    f"on the '{self.current_collection_name}' collection: {exc}"
                )
        if not processed_data:
            return rows
        if isinstance(processed_data, list):
            if not all(isinstance(item, dict) for item in processed_data):
                raise PostProcessingException(
                    "The list returned after postprocessing did not contain elements of the same type."
                )
            rows.extend(processed_data)
        elif isinstance(processed_data, dict):
            rows.append(processed_data)
        else:
            raise PostProcessingException(
                "Not enough information to continue processing. The result of postprocessing "
                f"must be an dict or a list of dicts, found value of '{processed_data}'"
            )

        return rows

    def mask_data(
        self,
        node: TraversalNode,
        policy: Policy,
        privacy_request: PrivacyRequest,
        rows: List[Row],
        input_data: Dict[str, List[Any]],
    ) -> int:
        """Execute a masking request. Return the number of rows that have been updated."""
        self.set_privacy_request_state(privacy_request, node)
        query_config = self.query_config(node)
        masking_request = query_config.get_masking_request()
        if not masking_request:
            raise Exception(
                f"Either no masking request configured or no valid masking request for {node.address.collection}. "
                f"Check that MASKING_STRICT env var is appropriately set"
            )

        self.set_saas_request_state(masking_request)

        # hook for user-provided request override functions
        if masking_request.request_override:
            return self._invoke_masking_request_override(
                masking_request.request_override,
                self.create_client(),
                policy,
                privacy_request,
                rows,
                query_config,
                masking_request,
                self.secrets,
            )

        # unwrap response using data_path
        if masking_request.data_path and rows:
            unwrapped = []
            for row in rows:
                unwrapped.extend(pydash.get(row, masking_request.data_path))
            rows = unwrapped

        # post-process access request response specific to masking request needs
        rows = self.process_response_data(
            rows,
            privacy_request.get_cached_identity_data(),
            cast(Optional[List[PostProcessorStrategy]], masking_request.postprocessors),
        )

        rows_updated = 0
        client = self.create_client()
        for row in rows:
            try:
                prepared_request = query_config.generate_update_stmt(
                    row, policy, privacy_request
                )
            except ValueError as exc:
                if masking_request.skip_missing_param_values:
                    logger.info(
                        "Skipping optional masking request on node {}: {}",
                        node.address.value,
                        exc,
                    )
                    continue
                raise exc
            client.send(prepared_request, masking_request.ignore_errors)
            rows_updated += 1

        self.unset_connector_state()
        return rows_updated

    def run_consent_request(
        self,
        node: TraversalNode,
        policy: Policy,
        privacy_request: PrivacyRequest,
        identity_data: Dict[str, Any],
        consent_preferences: List[Consent],
    ) -> bool:
        """Execute a consent request. Return whether the consent request to the third party succeeded.
        Return True if 200 OK
        """
        logger.info(
            "Starting consent request for node: '{}'",
            node.address.value,
        )
        self.set_privacy_request_state(privacy_request, node)
        query_config = self.query_config(node)

        for consent_options in consent_preferences:
            query_config.action = (
                "opt_in" if consent_options.opt_in else "opt_out"
            )  # For logging purposes

            consent_requests: List[
                SaaSRequest
            ] = self._get_consent_requests_by_preference(consent_options.opt_in)

            if not consent_requests:
                logger.info(
                    "Skipping consent requests on node {}: No '{}' requests defined",
                    node.address.value,
                    query_config.action,
                )
                continue

            for consent_request in consent_requests:
                self.set_saas_request_state(consent_request)

                try:
                    prepared_request: SaaSRequestParams = (
                        query_config.generate_consent_stmt(
                            policy, privacy_request, consent_request
                        )
                    )
                except ValueError as exc:
                    if consent_request.skip_missing_param_values:
                        logger.info(
                            "Skipping optional consent request on node {}: {}",
                            node.address.value,
                            exc,
                        )
                        continue
                    raise exc
                client: AuthenticatedClient = self.create_client()
                client.send(prepared_request)
        self.unset_connector_state()
        return True

    def close(self) -> None:
        """Not required for this type"""

    @staticmethod
    def _handle_errored_response(
        saas_request: SaaSRequest, response: Response
    ) -> Response:
        """
        Checks if given Response is an error and if SaasRequest is configured to ignore errors.
        If so, replaces given Response with empty dictionary.
        """
        if saas_request.ignore_errors and not response.ok:
            logger.info(
                "Ignoring and clearing errored response with status code {}.",
                response.status_code,
            )
            response = Response()
            response._content = b"{}"  # pylint: disable=W0212
        return response

    @staticmethod
    def _unwrap_response_data(saas_request: SaaSRequest, response: Response) -> Any:
        """
        Unwrap given Response using data_path in the given SaasRequest
        """
        try:
            return (
                pydash.get(response.json(), saas_request.data_path)
                if saas_request.data_path
                else response.json()
            )
        except JSONDecodeError:
            raise FidesopsException(
                f"Unable to parse JSON response from {saas_request.path}"
            )

    @staticmethod
    def _invoke_read_request_override(
        override_function_name: str,
        client: AuthenticatedClient,
        policy: Policy,
        privacy_request: PrivacyRequest,
        node: TraversalNode,
        input_data: Dict[str, List],
        secrets: Any,
    ) -> List[Row]:
        """
        Invokes the appropriate user-defined SaaS request override for read requests.

        Contains error handling for uncaught exceptions coming out of the override.
        """
        override_function: Callable[
            ..., Union[List[Row], int]
        ] = SaaSRequestOverrideFactory.get_override(
            override_function_name, SaaSRequestType.READ
        )
        try:
            return override_function(
                client,
                node,
                policy,
                privacy_request,
                input_data,
                secrets,
            )  # type: ignore
        except Exception as exc:
            logger.error(
                "Encountered error executing override access function '{}'",
                override_function_name,
                exc_info=True,
            )
            raise FidesopsException(str(exc))

    @staticmethod
    def _invoke_masking_request_override(
        override_function_name: str,
        client: AuthenticatedClient,
        policy: Policy,
        privacy_request: PrivacyRequest,
        rows: List[Row],
        query_config: SaaSQueryConfig,
        masking_request: SaaSRequest,
        secrets: Any,
    ) -> int:
        """
        Invokes the appropriate user-defined SaaS request override for masking
        (update, delete, data_protection_request) requests.

        Includes the necessary data preparations for override input
        and has error handling for uncaught exceptions coming out of the override
        """
        override_function: Callable[
            ..., Union[List[Row], int]
        ] = SaaSRequestOverrideFactory.get_override(
            override_function_name, SaaSRequestType(query_config.action)
        )
        try:
            # if using a saas override, we still need to use the core framework
            # to generate updated (masked) parameter values, and pass these
            # into the overridden function
            update_param_values: List[Dict[str, Any]] = [
                query_config.generate_update_param_values(
                    row, policy, privacy_request, masking_request
                )
                for row in rows
            ]
            return override_function(
                client,
                update_param_values,
                policy,
                privacy_request,
                secrets,
            )  # type: ignore
        except Exception:
            logger.error(
                "Encountered error executing override mask function '{}",
                override_function_name,
                exc_info=True,
            )
            raise FidesopsException(
                f"Error executing override mask function '{override_function_name}'"
            )

    def _get_consent_requests_by_preference(self, opt_in: bool) -> List[SaaSRequest]:
        """Helper to either pull out the opt-in requests or the opt out requests that were defined."""
        consent_requests: Optional[
            ConsentRequestMap
        ] = self.saas_config.consent_requests

        if not consent_requests:
            return []

        return consent_requests.opt_in if opt_in else consent_requests.opt_out  # type: ignore



"""Add optional contact information for organization, system, and dataset

Revision ID: 4c3693c289d0
Revises: 7c851d8a102a
Create Date: 2022-01-21 15:56:53.519486

"""
import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql
from sqlalchemy.dialects.postgresql import BYTEA

# revision identifiers, used by Alembic.
revision = "4c3693c289d0"
down_revision = "7c851d8a102a"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.execute("CREATE EXTENSION IF NOT EXISTS pgcrypto;")

    op.add_column("datasets", sa.Column("joint_controller", BYTEA, nullable=True))
    op.drop_column("evaluations", "updated_at")
    op.drop_column("evaluations", "created_at")
    op.add_column("organizations", sa.Column("controller", BYTEA, nullable=True))
    op.add_column(
        "organizations", sa.Column("data_protection_officer", BYTEA, nullable=True)
    )
    op.add_column("organizations", sa.Column("representative", BYTEA, nullable=True))
    op.add_column("systems", sa.Column("joint_controller", BYTEA, nullable=True))
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("systems", "joint_controller")
    op.drop_column("organizations", "controller")
    op.drop_column("organizations", "data_protection_officer")
    op.drop_column("organizations", "representative")
    op.add_column(
        "evaluations",
        sa.Column(
            "created_at",
            postgresql.TIMESTAMP(timezone=True),
            server_default=sa.text("now()"),
            autoincrement=False,
            nullable=True,
        ),
    )
    op.add_column(
        "evaluations",
        sa.Column(
            "updated_at",
            postgresql.TIMESTAMP(timezone=True),
            server_default=sa.text("now()"),
            autoincrement=False,
            nullable=True,
        ),
    )
    op.drop_column("datasets", "joint_controller")
    # ### end Alembic commands ###



from typing import AsyncGenerator

from sqlalchemy import create_engine
from sqlalchemy.ext.asyncio import AsyncSession, create_async_engine
from sqlalchemy.orm import Session, sessionmaker

from fides.core.config import get_config

config = get_config()

engine = create_async_engine(
    config.database.async_database_uri,
    echo=False,
)
async_session = sessionmaker(engine, class_=AsyncSession, expire_on_commit=False)

sync_engine = create_engine(config.database.sync_database_uri, echo=False)
sync_session = sessionmaker(
    sync_engine,
    class_=Session,
    expire_on_commit=False,
    autocommit=False,
)


async def get_async_db() -> AsyncGenerator:
    """Return an async session generator for dependency injection into API endpoints"""
    async with async_session() as session:
        yield session



from typing import Any, Dict, List

import firebase_admin
from firebase_admin import App, auth, credentials
from firebase_admin.auth import UserNotFoundError, UserRecord
from loguru import logger

from fides.api.ops.common_exceptions import FidesopsException
from fides.api.ops.graph.traversal import TraversalNode
from fides.api.ops.models.policy import Policy
from fides.api.ops.models.privacy_request import PrivacyRequest
from fides.api.ops.service.connectors.saas.authenticated_client import (
    AuthenticatedClient,
)
from fides.api.ops.service.saas_request.saas_request_override_factory import (
    SaaSRequestType,
    register,
)
from fides.api.ops.util.collection_util import Row
from fides.api.ops.util.logger import Pii
from fides.api.ops.util.saas_util import get_identity


@register("firebase_auth_user_access", [SaaSRequestType.READ])
def firebase_auth_user_access(  # pylint: disable=R0914
    client: AuthenticatedClient,
    node: TraversalNode,
    policy: Policy,
    privacy_request: PrivacyRequest,
    input_data: Dict[str, List[Any]],
    secrets: Dict[str, Any],
) -> List[Row]:
    """
    SaaS Request Override function used to integrate access requests with
    Firebase Auth service. Specifically, it reads `User` records.
    """
    app = initialize_firebase(secrets)

    processed_data = []
    identity = get_identity(privacy_request)
    user: UserRecord
    if identity == "email":
        emails = input_data.get("email", [])
        for email in emails:
            try:
                user = auth.get_user_by_email(email, app=app)
                processed_data.append(_user_record_to_row(user))
            except UserNotFoundError:
                logger.warning(
                    f"Could not find user with email {Pii(email)} in firebase"
                )
    elif identity == "phone_number":
        phone_numbers = input_data.get("phone_number", [])
        for phone_number in phone_numbers:
            try:
                user = auth.get_user_by_phone_number(phone_number, app=app)
                processed_data.append(_user_record_to_row(user))
            except UserNotFoundError:
                logger.warning(
                    f"Could not find user with phone_number {Pii(phone_number)} in firebase"
                )
    else:
        raise FidesopsException(
            "Unsupported identity type for Firebase connector. Currently only `email` and `phone_number` are supported"
        )
    return processed_data


def _user_record_to_row(user: UserRecord) -> Row:
    """
    Translates a Firebase `UserRecord` to a Fides access request result `Row`
    """
    row = {"uid": user.uid}
    keys = [
        "email",
        "uid",
        "display_name",
        "phone_number",
        "photo_url",
        "disabled",
        "email_verified",
    ]
    for key in keys:
        value = getattr(user, key)
        if value is not None:
            row[key] = value

    if user.provider_data:
        pds = []
        for pd in user.provider_data:
            pd_keys = ["display_name", "provider_id", "email", "photo_url"]
            pd_to_add = {}
            for key in pd_keys:
                value = getattr(pd, key)
                if value is not None:
                    pd_to_add[key] = value
            pds.append(pd_to_add)
        row["provider_data"] = pds
    return row


@register("firebase_auth_user_update", [SaaSRequestType.UPDATE])
def firebase_auth_user_update(
    client: AuthenticatedClient,
    param_values_per_row: List[Dict[str, Any]],
    policy: Policy,
    privacy_request: PrivacyRequest,
    secrets: Dict[str, Any],
) -> int:
    """
    SaaS Request Override function used to integrate update requests with
    Firebase Auth service. Specifically, it updates `User` records.
    """
    app = initialize_firebase(secrets)
    rows_updated = 0
    # each update_params dict correspond to a record that needs to be updated
    for row_param_values in param_values_per_row:
        user: UserRecord = _retrieve_user_record(privacy_request, row_param_values, app)
        masked_fields = row_param_values["masked_object_fields"]
        email = masked_fields.get("email", user.email)
        display_name = masked_fields.get("display_name", user.display_name)
        phone_number = masked_fields.get("phone_number", user.phone_number)
        photo_url = masked_fields.get("photo_url", user.photo_url)
        disabled = masked_fields.get("disabled", user.disabled)
        email_verified = masked_fields.get("email_verified", user.email_verified)
        auth.update_user(
            user.uid,
            email=email,
            display_name=display_name,
            phone_number=phone_number,
            photo_url=photo_url,
            disabled=disabled,
            email_verified=email_verified,
            app=app,
        )
        rows_updated += 1
    return rows_updated


@register("firebase_auth_user_delete", [SaaSRequestType.DELETE])
def firebase_auth_user_delete(
    client: AuthenticatedClient,
    param_values_per_row: List[Dict[str, Any]],
    policy: Policy,
    privacy_request: PrivacyRequest,
    secrets: Dict[str, Any],
) -> int:
    """
    SaaS Request Override function used to integrate delete requests with
    Firebase Auth service. Specifically, it deletes `User` records.

    Currently, this function is not invoked by any default configuration -
    instead, Firebase erasures are done via updates, and the update function
    defined above. We've kept this function implementation here so that
    it can be easily leveraged without code changes and by a simple config update.
    """
    app = initialize_firebase(secrets)
    rows_updated = 0
    # each update_params dict correspond to a record that needs to be updated
    for row_param_values in param_values_per_row:
        user: UserRecord = _retrieve_user_record(privacy_request, row_param_values, app)
        auth.delete_user(user.uid, app=app)
        rows_updated += 1
    return rows_updated


def _retrieve_user_record(
    privacy_request: PrivacyRequest, row_param_values: Dict[str, Any], app: App
) -> UserRecord:
    """
    Utility that erasure functions can use to retrieve a Firebase `UserRecord`
    """
    identity = get_identity(privacy_request)
    if identity == "email":
        email = row_param_values.get("email", [])
        return auth.get_user_by_email(email, app=app)
    if identity == "phone_number":
        phone_number = row_param_values.get("phone_number", [])
        return auth.get_user_by_phone_number(phone_number, app=app)

    raise FidesopsException(
        "Unsupported identity type for Firebase connector. Currently only `email` and `phone_number` are supported"
    )


def initialize_firebase(secrets: Dict[str, Any]) -> App:
    """
    Initializes a Firebase "app" instance based on the given
    secrets `dict`. The value of the `project_id` key of the
    secrets dict is used for the `name` of the Firebase app.
    """
    creds = get_firebase_credentials(secrets)
    project_id = creds.project_id
    try:
        app = firebase_admin.get_app(project_id)
    except ValueError:
        app = firebase_admin.initialize_app(credential=creds, name=project_id)
    return app


def get_firebase_credentials(secrets: Dict[str, Any]) -> credentials.Certificate:
    """
    Utility function to create a Firebase `credentials.Certificate` instance
    based on the provided secrets `dict`.
    """
    credential_keys = [
        "type",
        "project_id",
        "private_key_id",
        "private_key",
        "client_email",
        "client_id",
        "auth_uri",
        "token_uri",
        "auth_provider_x509_cert_url",
        "client_x509_cert_url",
    ]
    creds = {key: secrets[key] for key in credential_keys}
    return credentials.Certificate(cert=creds)



"""This module defines the settings for everything related to the CLI."""
from typing import Dict, Optional

from fideslog.sdk.python.utils import FIDESCTL_CLI, generate_client_id
from pydantic import AnyHttpUrl, validator

from .fides_settings import FidesSettings

# pylint: disable=C0115,C0116, E0213
ENV_PREFIX = "FIDES__CLI__"


class CLISettings(FidesSettings):
    """Class used to store values from the 'cli' section of the config."""

    local_mode: bool = False
    analytics_id: str = generate_client_id(FIDESCTL_CLI)

    # These defaults are required to make connecting to
    # docker instances possible by default
    server_protocol: str = "http"
    server_host: str = "localhost"
    server_port: str = "8080"
    server_url: Optional[AnyHttpUrl]

    @validator("server_url", always=True)
    @classmethod
    def get_server_url(cls, value: str, values: Dict) -> str:
        "Create the server_url."
        host = values["server_host"]
        port = int(values["server_port"])
        protocol = values["server_protocol"]

        server_url = "{}://{}{}".format(
            protocol,
            host,
            f":{port}" if port else "",
        )

        return server_url

    @validator("analytics_id", always=True)
    def ensure_not_empty(cls, value: str) -> str:
        """
        Validate that the `analytics_id` is not `""`.
        """
        return value if value != "" else generate_client_id(FIDESCTL_CLI)

    class Config:
        env_prefix = ENV_PREFIX



"""privacy notice attributes

Revision ID: adad3be6af08
Revises: 7a94527643a1
Create Date: 2022-03-22 19:22:15.248311

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "adad3be6af08"
down_revision = "fdd28c39a679"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "data_subjects",
        sa.Column("rights", sa.JSON(), nullable=True),
    )
    op.add_column(
        "data_subjects",
        sa.Column("automated_decisions_or_profiling", sa.BOOLEAN(), nullable=True),
    )
    op.add_column(
        "data_uses", sa.Column("legitimate_interest", sa.BOOLEAN(), nullable=True)
    )
    op.add_column(
        "data_uses",
        sa.Column("legitimate_interest_impact_assessment", sa.String(), nullable=True),
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("data_uses", "legitimate_interest_impact_assessment")
    op.drop_column("data_uses", "legitimate_interest")
    op.drop_column("data_subjects", "automated_decisions_or_profiling")
    op.drop_column("data_subjects", "rights")
    # ### end Alembic commands ###



import secrets

from fastapi import Security
from loguru import logger

from fides.api.ops.api.v1.scope_registry import ENCRYPTION_EXEC
from fides.api.ops.api.v1.urn_registry import (
    DECRYPT_AES,
    ENCRYPT_AES,
    ENCRYPTION_KEY,
    V1_URL_PREFIX,
)
from fides.api.ops.schemas.encryption_request import (
    AesDecryptionRequest,
    AesDecryptionResponse,
    AesEncryptionRequest,
    AesEncryptionResponse,
)
from fides.api.ops.util.api_router import APIRouter
from fides.api.ops.util.encryption.aes_gcm_encryption_scheme import (
    decrypt as aes_gcm_decrypt,
)
from fides.api.ops.util.encryption.aes_gcm_encryption_scheme import (
    encrypt_verify_secret_length as aes_gcm_encrypt,
)
from fides.api.ops.util.oauth_util import verify_oauth_client
from fides.core.config import get_config
from fides.lib.cryptography import cryptographic_util
from fides.lib.cryptography.cryptographic_util import b64_str_to_bytes, bytes_to_b64_str

router = APIRouter(tags=["Encryption"], prefix=V1_URL_PREFIX)


CONFIG = get_config()


@router.get(
    ENCRYPTION_KEY,
    dependencies=[Security(verify_oauth_client, scopes=[ENCRYPTION_EXEC])],
    response_model=str,
)
def get_encryption_key() -> str:
    logger.info("Generating encryption key")
    return cryptographic_util.generate_secure_random_string(
        CONFIG.security.aes_encryption_key_length
    )


@router.put(
    ENCRYPT_AES,
    dependencies=[Security(verify_oauth_client, scopes=[ENCRYPTION_EXEC])],
    response_model=AesEncryptionResponse,
)
def aes_encrypt(encryption_request: AesEncryptionRequest) -> AesEncryptionResponse:
    logger.info("Starting AES Encryption")
    nonce: bytes = secrets.token_bytes(CONFIG.security.aes_gcm_nonce_length)

    encrypted_value: str = aes_gcm_encrypt(
        encryption_request.value,
        encryption_request.key,  # type: ignore
        nonce,
    )
    return AesEncryptionResponse(
        encrypted_value=encrypted_value, nonce=bytes_to_b64_str(nonce)
    )


@router.put(
    DECRYPT_AES,
    dependencies=[Security(verify_oauth_client, scopes=[ENCRYPTION_EXEC])],
    response_model=AesDecryptionResponse,
)
def aes_decrypt(decryption_request: AesDecryptionRequest) -> AesDecryptionResponse:
    logger.info("Starting AES Decryption")
    nonce: bytes = b64_str_to_bytes(decryption_request.nonce)

    decrypted_value: str = aes_gcm_decrypt(
        decryption_request.value,
        decryption_request.key.encode(CONFIG.security.encoding),
        nonce,
    )
    return AesDecryptionResponse(decrypted_value=decrypted_value)



import abc
from typing import Any, Dict, List, Type

from fideslang import FidesDatasetReference
from pydantic import BaseModel, Extra, Field, PrivateAttr, create_model, root_validator
from pydantic.fields import FieldInfo
from sqlalchemy.orm import Session

from fides.api.ops.common_exceptions import ValidationError
from fides.api.ops.models.datasetconfig import validate_dataset_reference
from fides.api.ops.schemas.connection_configuration.connection_secrets import (
    ConnectionConfigSecretsSchema,
)
from fides.api.ops.schemas.saas.saas_config import SaaSConfig


class SaaSSchema(BaseModel, abc.ABC):
    """
    Abstract base schema for updating SaaS connection configuration secrets.
    Fields are added during runtime based on the connector_params and any
    external_references in the passed in saas_config"""

    @root_validator
    @classmethod
    def required_components_supplied(  # type: ignore
        cls, values: Dict[str, Any]
    ) -> Dict[str, Any]:
        """Validate that the minimum required components have been supplied."""

        # check required components are present
        required_components = [
            name for name, attributes in cls.__fields__.items() if attributes.required
        ]
        min_fields_present = all(
            values.get(component) for component in required_components
        )
        if not min_fields_present:
            raise ValueError(
                f"{cls.__name__} must be supplied all of: [{', '.join(required_components)}]."  # type: ignore
            )

        # check the types and values are consistent with the option and multivalue fields
        for name, value in values.items():
            connector_param = cls.get_connector_param(name)
            if connector_param:
                options = connector_param.get("options")
                multiselect = connector_param.get("multiselect")

                if options:
                    if isinstance(value, str):
                        if value not in options:
                            raise ValueError(
                                f"'{name}' must be one of [{', '.join(options)}]"
                            )
                    elif isinstance(value, list):
                        if not multiselect:
                            raise ValueError(
                                "f'{name}' must be a single value when multiselect is not enabled, not a list"
                            )
                        invalid_options = [
                            entry for entry in value if entry not in options
                        ]
                        if invalid_options:
                            raise ValueError(
                                f"[{', '.join(invalid_options)}] are not valid options, '{name}' must be a list of values from [{', '.join(options)}]"
                            )

        return values

    @classmethod
    def get_connector_param(cls, name: str) -> Dict[str, Any]:
        return cls.__private_attributes__.get("_connector_params").default.get(name)  # type: ignore

    @classmethod
    def external_references(cls) -> List[str]:
        return [
            name
            for name, property in cls.schema()["properties"].items()
            if "external_reference" in property and property["external_reference"]
        ]

    class Config:
        """Only permit selected secret fields to be stored."""

        extra = Extra.forbid
        orm_mode = True


class SaaSSchemaFactory:
    """
    Takes the base SaaS schema which only contains a field validator
    and uses the saas_config to generate the schema for the specific connector
    """

    def __init__(self, saas_config: SaaSConfig):
        self.saas_config = saas_config

    # Pydantic uses the shorthand of (type, ...) to denote a required field of the given type
    def get_saas_schema(self) -> Type[SaaSSchema]:
        """Returns the schema for the current configuration"""
        field_definitions: Dict[str, Any] = {}
        for connector_param in self.saas_config.connector_params:
            param_type = list if connector_param.multiselect else str
            field_definitions[connector_param.name] = (
                Field(
                    title=connector_param.label,
                    description=connector_param.description,
                    default=connector_param.default_value,
                )
                if connector_param.default_value
                else (
                    param_type,
                    FieldInfo(
                        title=connector_param.label,
                        description=connector_param.description,
                    ),
                )
            )
        if self.saas_config.external_references:
            for external_reference in self.saas_config.external_references:
                field_definitions[external_reference.name] = (
                    FidesDatasetReference,
                    FieldInfo(
                        title=external_reference.label,
                        description=external_reference.description,
                        external_reference=True,  # metadata added so we can identify these secret schema fields as external references
                    ),
                )
        SaaSSchema.__doc__ = f"{str(self.saas_config.type).capitalize()} secrets schema"  # Dynamically override the docstring to create a description

        # set the connector_params as a private attribute on the schema class
        # so they can be accessible in the 'required_components_supplied' validator
        model: Type[SaaSSchema] = create_model(
            f"{self.saas_config.type}_schema",
            **field_definitions,
            __base__=SaaSSchema,
            _connector_params=PrivateAttr(
                {
                    connector_param.name: {
                        "options": connector_param.options,
                        "multiselect": connector_param.multiselect,
                    }
                    for connector_param in self.saas_config.connector_params
                }
            ),
            _external_references=PrivateAttr(
                [
                    external_reference.name
                    for external_reference in self.saas_config.external_references
                ]
                if self.saas_config.external_references
                else []
            ),
        )

        return model


def validate_saas_secrets_external_references(
    db: Session,
    schema: SaaSSchema,
    connection_secrets: ConnectionConfigSecretsSchema,
) -> None:
    external_references = schema.external_references()
    for external_reference in external_references:
        dataset_reference: FidesDatasetReference = getattr(
            connection_secrets, external_reference
        )
        if dataset_reference.direction == "to":
            raise ValidationError(
                "External references can only have a direction of 'from', found 'to'"
            )
        validate_dataset_reference(db, dataset_reference)



"""This module defines the base settings for the other settings to inherit from."""

# pylint: disable=C0115,C0116, E0213

from typing import Tuple

from pydantic import BaseSettings, Extra
from pydantic.env_settings import SettingsSourceCallable


class FidesSettings(BaseSettings):
    """Class used as a base model for configuration subsections."""

    class Config:
        # Need to allow extras because the inheriting class will have more info
        extra = Extra.allow

        @classmethod
        def customise_sources(
            cls,
            init_settings: SettingsSourceCallable,
            env_settings: SettingsSourceCallable,
            file_secret_settings: SettingsSourceCallable,
        ) -> Tuple[SettingsSourceCallable, ...]:
            """Set environment variables to take precedence over init values."""
            return env_settings, init_settings, file_secret_settings



import copy
from collections import defaultdict
from typing import Any, Dict, List

import pydash
from loguru import logger

from fides.api.ops.task.refine_target_path import (
    DetailedPath,
    FieldPathNodeInput,
    build_refined_target_paths,
    join_detailed_path,
)
from fides.api.ops.util.collection_util import FIDESOPS_DO_NOT_MASK_INDEX, Row
from fides.api.ops.util.logger import Pii


def filter_element_match(
    row: Row,
    query_paths: FieldPathNodeInput,
    delete_elements: bool = True,
) -> Dict[str, Any]:
    """
    Modifies row in place to remove unmatched array elements or unmatched embedded documents within arrays.

    :param row: Record retrieved from a dataset
    :param query_paths: FieldPaths mapped to a list of values you want to match.
    :param delete_elements: If True, *removes* unmatched indices from array. If False, just *replaces* the data,
    so the original indices are preserved.
    :return: A modified record with array elements potentially eliminated if array data was targeted by a query path

    :Example:
    The given row was retrieved from a dataset because an element in row["A"] matched 2 or an element in
    row["C"]["D"] matched 5.  row["A"] is filtered to just contain the matching element, and row["C"] is filtered
    to just contain the objects where "D" = 5. Non-array elements should not be touched.

    filter_element_match(
        row={"A": [1, 2, 3], "B": 2, "C": [{"D": 3, "E": 4}, {"D": 5, "E": 6}, {"D": 5, "E": 7}]},
        query_paths={FieldPath("A"): [2], FieldPath("C, "D"): [5]}
    )

    {"A": [2], "B": 2, "C": [{"D": 5, "E": 6}, {"D": 5, "E": 7}]}

    :Example:
    With delete_elements=False, we replace elements with placeholder text instead.
    filter_element_match(
        row={"A": [1, 2, 3], "B": 2, "C": [{"D": 3, "E": 4}, {"D": 5, "E": 6}, {"D": 5, "E": 7}]},
        query_paths={FieldPath("A"): [2], FieldPath("C, "D"): [5]},
        delete_elements=False
    )

    {
     'A': ['FIDESOPS_DO_NOT_MASK', 2, 'FIDESOPS_DO_NOT_MASK'],
     'B': 2,
     'C': ['FIDESOPS_DO_NOT_MASK', {'D': 5, 'E': 6}, {'D': 5, 'E': 7}]
    }
    """
    detailed_target_paths: List[DetailedPath] = build_refined_target_paths(
        row, query_paths
    )

    array_paths_to_preserve: Dict[str, List[int]] = _expand_array_paths_to_preserve(
        detailed_target_paths
    )

    return _remove_paths_from_row(row, array_paths_to_preserve, delete_elements)


def _remove_paths_from_row(
    row: Dict[str, Any],
    preserve_indices: Dict[str, List[int]],
    delete_elements: bool = True,
) -> Dict[str, Any]:
    """
    Used by filter_element_match, remove array elements from row that are not specified in preserve_indices

    :param row: Record retrieved from a dataset
    :param preserve_indices: A dictionary of dot-separated paths to arrays, where the values are the list of indices
     we want to *keep*
    :param delete_elements: True if we just want to remove the element at the given index, otherwise we replace
     the element with some placeholder text.
    :return: A filtered row that has removed non-specified indices from arrays

    :Example:
    The first element in row["A"]["B"] was the only one specified to preserve, so we remove the other two.
    _remove_paths_from_row({"A": {"B": [{"C": "D"}, {"C": "F"}, {"C": "G"}]}}, {"A.B": [0]})

    {'A': {'B': [{'C': 'D'}]}}
    """
    desc_path_length: Dict[
        str, List[int]
    ] = dict(  # We want to remove elements from the deepest paths first
        sorted(
            preserve_indices.items(),
            key=lambda item: item[0].count("."),
            reverse=True,
        )
    )
    for path, preserve_index_list in desc_path_length.items():
        matched_array: List = pydash.objects.get(row, path)
        if matched_array is None:
            # This case shouldn't happen - if this gets logged, we've done something wrong
            logger.info(
                "_remove_paths_from_row call: Path {} in row {} not found.",
                path,
                Pii(row),
            )
            continue
        # Loop through array in *reverse* to delete/replace indices
        for index, _ in reversed(list(enumerate(matched_array))):
            if index not in preserve_index_list:  # List of indices that we want to keep
                if delete_elements:  # We delete the element at the given index
                    matched_array.pop(index)
                else:  # We replace the element at the given index
                    matched_array[index] = FIDESOPS_DO_NOT_MASK_INDEX

    return row


def _expand_array_paths_to_preserve(paths: List[DetailedPath]) -> Dict[str, List[int]]:
    """
    Used by "filter_element_match" - Returns a dictionary of string paths mapped to array indices that we want
    to preserve.

    :param paths: A list of lists of detailed paths (containing strings and array indices) to elements that matched query criteria
    :return: A dict where the keys are a dot-separated path to an array, and the values are a list of indices
    in that array that we want to keep.  If there are no indices in the original path, that path will be ignored.
    Some paths may be expanded into multiple paths where there are multiple levels of indices (arrays of arrays).

    :Example:
    _expand_array_paths_to_preserve([["F", 1, 2], ["F", 1, 3], ["G", "H"], ["L", 1, "M"]])
    {'F': [1], 'F.1': [2, 3], 'L': [1]}

    This data will be used to remove all elements from row["F"][1] that are not at index 2, and 3.
    We'll then remove all elements from "F" that are not at index [1], and all elements from "L" that are not at index 1.

    """
    # Break path into multiple paths if array elements in path
    expanded: List[DetailedPath] = []
    for path in paths:
        while path != [] and not isinstance(path[-1], int):
            path.pop()
        new_path: DetailedPath = []
        for elem in path:
            new_path.append(elem)
            if isinstance(elem, int) and new_path not in expanded:
                expanded.append(copy.deepcopy(new_path))

    # Combine paths where the key is a dot-separated path to the array, and the value are the indices
    # of the array we want to preserve
    merge_paths: Dict[str, List[int]] = defaultdict(list)
    for path in expanded:
        merge_paths[join_detailed_path(path[0:-1])].append(path[-1])  # type: ignore
    return merge_paths



from .fides_settings import FidesSettings

ENV_PREFIX = "FIDES__EXECUTION__"


class ExecutionSettings(FidesSettings):
    """Configuration settings for execution."""

    privacy_request_delay_timeout: int = 3600
    task_retry_count: int = 0
    task_retry_delay: int = 1  # In seconds
    task_retry_backoff: int = 1
    subject_identity_verification_required: bool = False
    require_manual_request_approval: bool = False
    masking_strict: bool = True

    class Config:
        env_prefix = ENV_PREFIX



from __future__ import annotations

from asyncio import sleep
from datetime import datetime
from typing import Any, Dict, List, Optional, Set

from httpx import AsyncClient
from loguru import logger

from fides.api.ops.api.v1.urn_registry import PRIVACY_REQUESTS, V1_URL_PREFIX
from fides.api.ops.common_exceptions import PrivacyRequestNotFound
from fides.api.ops.models.policy import ActionType, Policy
from fides.api.ops.models.privacy_request import PrivacyRequest, PrivacyRequestStatus
from fides.api.ops.schemas.drp_privacy_request import DrpPrivacyRequestCreate
from fides.api.ops.schemas.masking.masking_secrets import MaskingSecretCache
from fides.api.ops.schemas.privacy_request import PrivacyRequestResponse
from fides.api.ops.schemas.redis_cache import Identity
from fides.api.ops.service.masking.strategy.masking_strategy import MaskingStrategy
from fides.core.config import get_config

CONFIG = get_config()


def build_required_privacy_request_kwargs(
    requested_at: Optional[datetime], policy_id: str
) -> Dict[str, Any]:
    """Build kwargs required for creating privacy request

    If identity verification is on, the request will have an initial status of
    "identity_unverified", otherwise, it will have an initial status of "pending".
    """
    status = (
        PrivacyRequestStatus.identity_unverified
        if CONFIG.execution.subject_identity_verification_required
        else PrivacyRequestStatus.pending
    )
    return {
        "requested_at": requested_at,
        "policy_id": policy_id,
        "status": status,
    }


def cache_data(
    privacy_request: PrivacyRequest,
    policy: Policy,
    identity: Identity,
    encryption_key: Optional[str],
    drp_request_body: Optional[DrpPrivacyRequestCreate],
) -> None:
    """Cache privacy request data"""
    # Store identity and encryption key in the cache
    logger.info("Caching identity for privacy request {}", privacy_request.id)
    privacy_request.cache_identity(identity)
    privacy_request.cache_encryption(encryption_key)  # handles None already

    # Store masking secrets in the cache
    logger.info("Caching masking secrets for privacy request {}", privacy_request.id)
    erasure_rules = policy.get_rules_for_action(action_type=ActionType.erasure)
    unique_masking_strategies_by_name: Set[str] = set()
    for rule in erasure_rules:
        strategy_name: str = rule.masking_strategy["strategy"]  # type: ignore
        configuration = rule.masking_strategy["configuration"]  # type: ignore
        if strategy_name in unique_masking_strategies_by_name:
            continue
        unique_masking_strategies_by_name.add(strategy_name)
        masking_strategy = MaskingStrategy.get_strategy(strategy_name, configuration)
        if masking_strategy.secrets_required():
            masking_secrets: List[
                MaskingSecretCache
            ] = masking_strategy.generate_secrets_for_cache()
            for masking_secret in masking_secrets:
                privacy_request.cache_masking_secret(masking_secret)
    if drp_request_body:
        privacy_request.cache_drp_request_body(drp_request_body)


def get_async_client() -> AsyncClient:
    """Return an async client used to make API requests"""
    return AsyncClient()


async def poll_server_for_completion(
    privacy_request_id: str,
    server_url: str,
    token: str,
    *,
    poll_interval_seconds: int = 30,
    timeout_seconds: int = 1800,  # 30 minutes
    client: AsyncClient | None = None,
) -> PrivacyRequestResponse:
    """Poll a server for privacy request completion.

    Requests will report complete with if they have a status of canceled, complete,
    denied, or error. By default the polling will time out if not completed in 30
    minutes, time can be overridden by setting the timeout_seconds.
    """
    url = (
        f"{server_url}{V1_URL_PREFIX}{PRIVACY_REQUESTS}?request_id={privacy_request_id}"
    )
    start_time = datetime.now()
    elapsed_time = 0.0
    while elapsed_time < timeout_seconds:
        if client:
            response = await client.get(
                url, headers={"Authorization": f"Bearer {token}"}
            )
        else:
            async_client = get_async_client()
            response = await async_client.get(
                url, headers={"Authorization": f"Bearer {token}"}
            )
        response.raise_for_status()

        # Privacy requests are returned paginated. Since this is searching for a specific
        # privacy request there should only be one value present in items.
        items = response.json()["items"]
        if not items:
            raise PrivacyRequestNotFound(
                f"No privacy request found with id '{privacy_request_id}'"
            )
        status = PrivacyRequestResponse(**items[0])
        if status.status and status.status in (
            PrivacyRequestStatus.complete,
            PrivacyRequestStatus.canceled,
            PrivacyRequestStatus.error,
            PrivacyRequestStatus.denied,
        ):
            return status

        await sleep(poll_interval_seconds)
        time_delta = datetime.now() - start_time
        elapsed_time = time_delta.seconds
    raise TimeoutError(
        f"Timeout of {timeout_seconds} seconds has been exceeded while waiting for privacy request {privacy_request_id}"
    )



from typing import Optional

from fastapi import Depends, Security
from fastapi_pagination import Page, Params
from fastapi_pagination.bases import AbstractPage
from fastapi_pagination.ext.sqlalchemy import paginate
from fideslang.validation import FidesKey
from loguru import logger
from sqlalchemy.orm import Session
from starlette.exceptions import HTTPException
from starlette.status import (
    HTTP_200_OK,
    HTTP_204_NO_CONTENT,
    HTTP_400_BAD_REQUEST,
    HTTP_404_NOT_FOUND,
    HTTP_422_UNPROCESSABLE_ENTITY,
    HTTP_500_INTERNAL_SERVER_ERROR,
)

from fides.api.ops.api import deps
from fides.api.ops.api.v1.scope_registry import (
    MESSAGING_CREATE_OR_UPDATE,
    MESSAGING_DELETE,
    MESSAGING_READ,
)
from fides.api.ops.api.v1.urn_registry import (
    MESSAGING_BY_KEY,
    MESSAGING_CONFIG,
    MESSAGING_SECRETS,
    V1_URL_PREFIX,
)
from fides.api.ops.common_exceptions import MessagingConfigNotFoundException
from fides.api.ops.models.messaging import MessagingConfig, get_schema_for_secrets
from fides.api.ops.schemas.messaging.messaging import (
    MessagingConfigRequest,
    MessagingConfigResponse,
    TestMessagingStatusMessage,
)
from fides.api.ops.schemas.messaging.messaging_secrets_docs_only import (
    possible_messaging_secrets,
)
from fides.api.ops.service.messaging.messaging_crud_service import (
    create_or_update_messaging_config,
    delete_messaging_config,
    get_messaging_config_by_key,
    update_messaging_config,
)
from fides.api.ops.util.api_router import APIRouter
from fides.api.ops.util.logger import Pii
from fides.api.ops.util.oauth_util import verify_oauth_client

router = APIRouter(tags=["messaging"], prefix=V1_URL_PREFIX)


@router.post(
    MESSAGING_CONFIG,
    status_code=HTTP_200_OK,
    dependencies=[Security(verify_oauth_client, scopes=[MESSAGING_CREATE_OR_UPDATE])],
    response_model=MessagingConfigResponse,
)
def post_config(
    *,
    db: Session = Depends(deps.get_db),
    messaging_config: MessagingConfigRequest,
) -> MessagingConfigResponse:
    """
    Given a messaging config, create corresponding MessagingConfig object, provided no config already exists
    """

    try:
        return create_or_update_messaging_config(db=db, config=messaging_config)
    except ValueError as e:
        logger.warning(
            "Create failed for messaging config {}: {}",
            messaging_config.key,
            Pii(str(e)),
        )
        raise HTTPException(
            status_code=HTTP_400_BAD_REQUEST,
            detail=f"Config with key {messaging_config.key} failed to be added: {e}",
        )
    except Exception as exc:
        logger.warning(
            "Create failed for messaging config {}: {}",
            messaging_config.key,
            Pii(str(exc)),
        )
        raise HTTPException(
            status_code=HTTP_500_INTERNAL_SERVER_ERROR,
            detail=f"Config with key {messaging_config.key} failed to be added: {exc}",
        )


@router.patch(
    MESSAGING_BY_KEY,
    dependencies=[Security(verify_oauth_client, scopes=[MESSAGING_CREATE_OR_UPDATE])],
    response_model=MessagingConfigResponse,
)
def patch_config_by_key(
    config_key: FidesKey,
    *,
    db: Session = Depends(deps.get_db),
    messaging_config: MessagingConfigRequest,
) -> Optional[MessagingConfigResponse]:
    """
    Updates config for messaging by key, provided config with key can be found.
    """
    try:
        return update_messaging_config(db=db, key=config_key, config=messaging_config)
    except MessagingConfigNotFoundException:
        logger.warning("No messaging config found with key {}", config_key)
        raise HTTPException(
            status_code=HTTP_404_NOT_FOUND,
            detail=f"No messaging config found with key {config_key}",
        )
    except Exception as exc:
        logger.warning(
            "Patch failed for messaging config {}: {}",
            messaging_config.key,
            Pii(str(exc)),
        )
        raise HTTPException(
            status_code=HTTP_500_INTERNAL_SERVER_ERROR,
            detail=f"Config with key {messaging_config.key} failed to be added",
        )


@router.put(
    MESSAGING_SECRETS,
    status_code=HTTP_200_OK,
    dependencies=[Security(verify_oauth_client, scopes=[MESSAGING_CREATE_OR_UPDATE])],
    response_model=TestMessagingStatusMessage,
)
def put_config_secrets(
    config_key: FidesKey,
    *,
    db: Session = Depends(deps.get_db),
    unvalidated_messaging_secrets: possible_messaging_secrets,
) -> TestMessagingStatusMessage:
    """
    Add or update secrets for messaging config.
    """
    logger.info("Finding messaging config with key '{}'", config_key)
    messaging_config = MessagingConfig.get_by(db=db, field="key", value=config_key)
    if not messaging_config:
        raise HTTPException(
            status_code=HTTP_404_NOT_FOUND,
            detail=f"No messaging configuration with key {config_key}.",
        )

    try:
        secrets_schema = get_schema_for_secrets(
            service_type=messaging_config.service_type,
            secrets=unvalidated_messaging_secrets,
        )
    except KeyError as exc:
        raise HTTPException(
            status_code=HTTP_422_UNPROCESSABLE_ENTITY,
            detail=exc.args[0],
        )
    except ValueError as exc:
        raise HTTPException(
            status_code=HTTP_400_BAD_REQUEST,
            detail=exc.args[0],
        )

    logger.info(
        "Updating messaging config secrets for config with key '{}'", config_key
    )
    try:
        messaging_config.set_secrets(db=db, messaging_secrets=secrets_schema.dict())
    except ValueError as exc:
        raise HTTPException(
            status_code=HTTP_400_BAD_REQUEST,
            detail=exc.args[0],
        )
    msg = f"Secrets updated for MessagingConfig with key: {config_key}."
    # todo- implement test status for messaging service
    return TestMessagingStatusMessage(msg=msg, test_status=None)


@router.get(
    MESSAGING_CONFIG,
    dependencies=[Security(verify_oauth_client, scopes=[MESSAGING_READ])],
    response_model=Page[MessagingConfigResponse],
)
def get_configs(
    *, db: Session = Depends(deps.get_db), params: Params = Depends()
) -> AbstractPage[MessagingConfig]:
    """
    Retrieves configs for messaging.
    """
    logger.info(
        "Finding all messaging configurations with pagination params {}", params
    )
    return paginate(
        MessagingConfig.query(db=db).order_by(MessagingConfig.created_at.desc()),
        params=params,
    )


@router.get(
    MESSAGING_BY_KEY,
    dependencies=[Security(verify_oauth_client, scopes=[MESSAGING_READ])],
    response_model=MessagingConfigResponse,
)
def get_config_by_key(
    config_key: FidesKey, *, db: Session = Depends(deps.get_db)
) -> MessagingConfigResponse:
    """
    Retrieves configs for messaging service by key.
    """
    logger.info("Finding messaging config with key '{}'", config_key)

    try:
        return get_messaging_config_by_key(db=db, key=config_key)
    except MessagingConfigNotFoundException as e:
        raise HTTPException(
            status_code=HTTP_404_NOT_FOUND,
            detail=e.message,
        )


@router.delete(
    MESSAGING_BY_KEY,
    status_code=HTTP_204_NO_CONTENT,
    dependencies=[Security(verify_oauth_client, scopes=[MESSAGING_DELETE])],
)
def delete_config_by_key(
    config_key: FidesKey, *, db: Session = Depends(deps.get_db)
) -> None:
    """
    Deletes messaging configs by key.
    """
    try:
        delete_messaging_config(db=db, key=config_key)
    except MessagingConfigNotFoundException as e:
        raise HTTPException(
            status_code=HTTP_404_NOT_FOUND,
            detail=e.message,
        )
